"""
Base model used for products.  Stores hierarchical categories
as well as individual product level information which includes
options.
"""

import config
import datetime
import keyedcache
import logging
import random
import signals
import operator
import os.path

from decimal import Decimal
from django import forms
from django.conf import settings
from django.contrib.sites.models import Site
from django.core import urlresolvers
from django.db import models
from django.db.models import Q
from django.db.models.fields.files import FileField
from django.utils.encoding import smart_str

#from django.utils.hashcompat import sha_constructor
from hashlib import sha1
from django.utils.safestring import mark_safe
from django.utils.translation import get_language, ugettext, ugettext_lazy as _
from l10n.utils import moneyfmt
from livesettings import config_value, SettingNotSet, config_value_safe
from satchmo_utils import cross_list, normalize_dir, url_join, get_flat_list, add_month
from satchmo_utils.fields import CurrencyField
from satchmo_utils.thumbnail.field import ImageWithThumbnailField
from satchmo_utils.unique_id import slugify

log = logging.getLogger('product.models')

dimension_units = (('cm','cm'), ('in','in'))

weight_units = (('kg','kg'), ('lb','lb'))

DISCOUNT_SHIPPING_CHOICES = (
    ('NONE', _('None')),
    ('FREE', _('Free Shipping')),
    ('FREECHEAP', _('Cheapest shipping option is free')),
    ('APPLY', _('Apply the discount above to shipping'))
)

SHIP_CLASS_CHOICES = (
    ('DEFAULT', _('Default')),
    ('YES', _('Shippable')),
    ('NO', _('Not Shippable'))
)

def default_dimension_unit():
    val = config_value_safe('PRODUCT','MEASUREMENT_SYSTEM', (None, None))[0]
    if val == 'metric':
        return 'cm'
    else:
        return 'in'

def default_weight_unit():
    val = config_value_safe('PRODUCT','MEASUREMENT_SYSTEM', (None, None))[0]
    if val == 'metric':
        return 'kg'
    else:
        return 'lb'

class CategoryManager(models.Manager):
    def active(self):
        return self.filter(is_active=True)
    
    def by_site(self, site=None, **kwargs):
        """Get all categories for this site"""
        if not site:
            site = Site.objects.get_current()
        
        site = site.id

        return self.active().filter(site__id__exact = site, **kwargs)
        
    def get_by_site(self, site=None, **kwargs):
        if not site:
            site = Site.objects.get_current()
        return self.active().get(site = site, **kwargs)
    
    def root_categories(self, site=None, **kwargs):
        """Get all root categories."""
        
        if not site:
            site = Site.objects.get_current()
        
        return self.active().filter(parent__isnull=True, site=site, **kwargs)
        
    def search_by_site(self, keyword, site=None, include_children=False):
        """Search for categories by keyword. 
        Note, this does not return a queryset."""
        
        if not site:
            site = Site.objects.get_current()
        
        cats = self.active().filter(
            Q(name__icontains=keyword) |
            Q(meta__icontains=keyword) |
            Q(description__icontains=keyword),
            site=site)
        
        if include_children:
            # get all the children of the categories found
            cats = [cat.get_active_children(include_self=True) for cat in cats]
            
        # sort properly
        if cats:
            fastsort = [(c.ordering, c.name, c) for c in get_flat_list(cats)]
            fastsort.sort()        
        if not site:
            site = Site.objects.get_current()
        
        cats = self.active().filter(
            Q(name__icontains=keyword) |
            Q(meta__icontains=keyword) |
            Q(description__icontains=keyword),
            site=site)
        
        if include_children:
            # get all the children of the categories found
            cats = [cat.get_active_children(include_self=True) for cat in cats]
            
        # sort properly
        if cats:
            fastsort = [(c.ordering, c.name, c) for c in get_flat_list(cats)]
            fastsort.sort()

            # extract the cat list
            cats = zip(*fastsort)[2]            
        return cats

class Category(models.Model):
    """
    Basic hierarchical category model for storing products
    """
    site = models.ForeignKey(Site, verbose_name=_('Site'))
    name = models.CharField(_("Name"), max_length=200)
    slug = models.SlugField(_("Slug"), help_text=_("Used for URLs, auto-generated from name if blank"), blank=True)
    parent = models.ForeignKey('self', blank=True, null=True,
        related_name='child')
        #,validator_list=['categoryvalidator'])
    meta = models.TextField(_("Meta Description"), blank=True, null=True,
        help_text=_("Meta description for this category"))
    description = models.TextField(_("Description"), blank=True,
        help_text="Optional")
    ordering = models.IntegerField(_("Ordering"), default=0, help_text=_("Override alphabetical order in category display"))
    is_active = models.BooleanField(default=True, blank=True)
    related_categories = models.ManyToManyField('self', blank=True, null=True, 
        verbose_name=_('Related Categories'), related_name='related_categories')
    objects = CategoryManager()

    def _get_mainImage(self):
        img = False
        if self.images.count() > 0:
            img = self.images.order_by('sort')[0]
        else:
            if self.parent_id and self.parent != self:
                img = self.parent.main_image

        if not img:
            #This should be a "Image Not Found" placeholder image
            try:
                img = CategoryImage.objects.filter(category__isnull=True).order_by('sort')[0]
            except IndexError:
                import sys
                print >>sys.stderr, 'Warning: default category image not found - try syncdb'

        return img

    main_image = property(_get_mainImage)

    def active_products(self, variations=True, include_children=False, **kwargs):
        if not include_children:
            qry = self.product_set.all()
        else:
            cats = self.get_all_children(include_self=True)
            qry = Product.objects.filter(category__in=cats)
            
        if variations:
            return qry.filter(site=self.site, active=True, **kwargs)
        else:
            return qry.filter(site=self.site, active=True, productvariation__parent__isnull=True, **kwargs)

    def translated_description(self, language_code=None):
        return lookup_translation(self, 'description', language_code)

    def translated_name(self, language_code=None):
        return lookup_translation(self, 'name', language_code)

    def _recurse_for_parents(self, cat_obj):
        p_list = []
        if cat_obj.parent_id:
            p = cat_obj.parent
            p_list.append(p)
            if p != self:
                more = self._recurse_for_parents(p)
                p_list.extend(more)
        if cat_obj == self and p_list:
            p_list.reverse()
        return p_list
        
    def parents(self):
        return self._recurse_for_parents(self)

    @models.permalink
    def get_absolute_url(self):
        parents = self._recurse_for_parents(self)
        slug_list = [cat.slug for cat in parents]
        if slug_list:
            slug_list = "/".join(slug_list) + "/"
        else:
            slug_list = ""
        return ('satchmo_category', (), {'parent_slugs' : slug_list, 
            'slug' : self.slug})

    def get_separator(self):
        return ' :: '

    def _parents_repr(self):
        name_list = [cat.name for cat in self._recurse_for_parents(self)]
        return self.get_separator().join(name_list)
    _parents_repr.short_description = "Category parents"

    def get_url_name(self):
        # Get all the absolute URLs and names for use in the site navigation.
        name_list = []
        url_list = []
        for cat in self._recurse_for_parents(self):
            name_list.append(cat.translated_name())
            url_list.append(cat.get_absolute_url())
        name_list.append(self.translated_name())
        url_list.append(self.get_absolute_url())
        return zip(name_list, url_list)

    def __unicode__(self):
        name_list = [cat.name for cat in self._recurse_for_parents(self)]
        name_list.append(self.name)
        return self.get_separator().join(name_list)

    def save(self, **kwargs):
        if self.id:
            if self.parent and self.parent_id == self.id:
                raise forms.ValidationError(_("You must not save a category in itself!"))

            for p in self._recurse_for_parents(self):
                if self.id == p.id:
                    raise forms.ValidationError(_("You must not save a category in itself!"))

        if not self.slug:
            self.slug = slugify(self.name, instance=self)

        super(Category, self).save(**kwargs)

    def _flatten(self, L):
        """
        Taken from a python newsgroup post
        """
        if type(L) != type([]): return [L]
        if L == []: return L
        return self._flatten(L[0]) + self._flatten(L[1:])

    def _recurse_for_children(self, node, only_active=False):
        children = []
        children.append(node)
        for child in node.child.active():
            if child != self:
                if (not only_active) or child.active_products().count() > 0:
                    children_list = self._recurse_for_children(child, only_active=only_active)
                    children.append(children_list)
        return children

    def get_active_children(self, include_self=False):
        """
        Gets a list of all of the children categories which have active products.
        """
        return self.get_all_children(only_active=True, include_self=include_self)

    def get_all_children(self, only_active=False, include_self=False):
        """
        Gets a list of all of the children categories.
        """
        children_list = self._recurse_for_children(self, only_active=only_active)
        if include_self:
            ix = 0
        else:
            ix = 1
        flat_list = self._flatten(children_list[ix:])
        return flat_list
        
    class Meta:
        ordering = ['site', 'parent__id', 'ordering', 'name']
        verbose_name = _("Category")
        verbose_name_plural = _("Categories")
        unique_together = ('site', 'slug')

class CategoryTranslation(models.Model):
    """A specific language translation for a `Category`.  This is intended for all descriptions which are not the
    default settings.LANGUAGE.
    """
    category = models.ForeignKey(Category, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    name = models.CharField(_("Translated Category Name"), max_length=255, )
    description = models.TextField(_("Description of category"), default='', blank=True)
    version = models.IntegerField(_('version'), default=1)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('Category Translation')
        verbose_name_plural = _('Category Translations')
        ordering = ('category', 'name','languagecode')
        unique_together = ('category', 'languagecode', 'version')

    def __unicode__(self):
        return u"CategoryTranslation: [%s] (ver #%i) %s Name: %s" % (self.languagecode, self.version, self.category, self.name)

class CategoryImage(models.Model):
    """
    A picture of an item.  Can have many pictures associated with an item.
    Thumbnails are automatically created.
    """
    category = models.ForeignKey(Category, null=True, blank=True,
        related_name="images")
    picture = ImageWithThumbnailField(verbose_name=_('Picture'),
        upload_to="__DYNAMIC__", # eek...
        name_field="_filename",
        max_length=200,
        null=True,
        blank=True) #Media root is automatically prepended
    caption = models.CharField(_("Optional caption"), max_length=100,
        null=True, blank=True)
    sort = models.IntegerField(_("Sort Order"), default=0)

    def translated_caption(self, language_code=None):
        return lookup_translation(self, 'caption', language_code)

    def _get_filename(self):
        if self.category:
            return '%s-%s' % (self.category.slug, self.id)
        else:
            return 'default'
    _filename = property(_get_filename)

    def __unicode__(self):
        if self.category:
            return u"Image of Category %s" % self.category.slug
        elif self.caption:
            return u"Image with caption \"%s\"" % self.caption
        else:
            return u"%s" % self.picture

    class Meta:
        ordering = ['sort']
        unique_together = (('category', 'sort'),)
        verbose_name = _("Category Image")
        verbose_name_plural = _("Category Images")

class CategoryImageTranslation(models.Model):
    """A specific language translation for a `CategoryImage`.  This is intended for all descriptions which are not the
    default settings.LANGUAGE.
    """
    categoryimage = models.ForeignKey(CategoryImage, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    caption = models.CharField(_("Translated Caption"), max_length=255, )
    version = models.IntegerField(_('version'), default=1)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('Category Image Translation')
        verbose_name_plural = _('Category Image Translations')
        ordering = ('categoryimage', 'caption','languagecode')
        unique_together = ('categoryimage', 'languagecode', 'version')

    def __unicode__(self):
        return u"CategoryImageTranslation: [%s] (ver #%i) %s Name: %s" % (self.languagecode, self.version, self.categoryimage, self.name)

class OptionGroupManager(models.Manager):
    def get_sortmap(self):
        """Returns a dictionary mapping ids to sort order"""
        
        work = {}
        for uid, order in self.values_list('id', 'sort_order'):
            work[uid] = order

        return work

class NullDiscount(object):

    def __init__(self):
        self.description = _("No Discount")
        self.total = Decimal("0.00")
        self.item_discounts = {}
        self.discounted_prices = []
        self.automatic = False

    def calc(self, *args):
        return Decimal("0.00")
        
    def is_valid(self):
        return False
        
    def valid_for_product(self, product):
        return False


class DiscountManager(models.Manager):
    def by_code(self, code, raises=False):
        discount = None

        if code:
            try:
                return self.get(code=code, active=True)

            except Discount.DoesNotExist:
                if raises:
                    raise
                    
        return NullDiscount()
        
    def get_sale(self):
        """Get the current 'sale' discount."""
        today = datetime.date.today()
        sale = None
        site = Site.objects.get_current()
        try:
            sale = keyedcache.cache_get('discount', 'sale', site, today)
        except keyedcache.NotCachedError, nce:
            discs = self.filter(automatic=True, 
                active=True, 
                site=site,
                startDate__lte=today, 
                endDate__gt=today).order_by('-percentage')
            if discs.count() > 0:
                sale = discs[0]

            keyedcache.cache_set(nce.key, value=sale)
            
        if sale is None:
            raise Discount.DoesNotExist
        else:
            return sale

class Discount(models.Model):
    """
    Allows for multiple types of discounts including % and dollar off.
    Also allows finite number of uses.
    """
    site = models.ForeignKey(Site, verbose_name=_('site'))
    description = models.CharField(_("Description"), max_length=100)
    code = models.CharField(_("Discount Code"), max_length=20, unique=True,
        help_text=_("Coupon Code"))
    active = models.BooleanField(_("Active"))
    amount = CurrencyField(_("Discount Amount"), decimal_places=2,
        max_digits=8, blank=True, null=True, 
        #validator_list=[amount_validator],
        help_text=_("Enter absolute discount amount OR percentage."))
    percentage = models.DecimalField(_("Discount Percentage"), decimal_places=2,
        max_digits=5, blank=True, null=True,
        #validator_list=[percentage_validator],
        help_text=_("Enter absolute discount amount OR percentage.  Percents are given in whole numbers, and can be up to 100%."))
    automatic = models.NullBooleanField(_("Is this an automatic discount?"), default=False, blank=True,
        null=True, help_text=_("Use this field to advertise the discount on all products to which it applies.  Generally this is used for site-wide sales."))
    allowedUses = models.IntegerField(_("Number of allowed uses"),
        blank=True, null=True, help_text=_('Set this to a number greater than 0 to have the discount expire after that many uses.'))
    numUses = models.IntegerField(_("Number of times already used"),
        blank=True, null=True)
    minOrder = CurrencyField(_("Minimum order value"),
        decimal_places=2, max_digits=8, blank=True, null=True)
    startDate = models.DateField(_("Start Date"))
    endDate = models.DateField(_("End Date"))
    shipping = models.CharField(_("Shipping"), choices=DISCOUNT_SHIPPING_CHOICES, 
        default='NONE', blank=True, null=True, max_length=10)
    allValid = models.BooleanField(_("All products?"), default=False, 
        help_text=_('Apply this discount to all discountable products? If this is false you must select products below in the "Valid Products" section.'))
    validProducts = models.ManyToManyField('Product', verbose_name=_("Valid Products"),
        blank=True, null=True)

    objects = DiscountManager()
        
    def __init__(self, *args, **kwargs):
        self._calculated = False
        super(Discount, self).__init__(*args, **kwargs)

    def __unicode__(self):
        return self.description

    def isValid(self, cart=None, contact=None):
        """
        Make sure this discount still has available uses and is in the current date range.
        If a cart has been populated, validate that it does apply to the products we have selected.
        If this is a "FREECHEAP" discount, then error if the cheapest shipping hasn't been chosen.
        """
        if not self.active:
            return (False, ugettext('This coupon is disabled.'))
        if self.startDate > datetime.date.today():
            return (False, ugettext('This coupon is not active yet.'))
        if self.endDate < datetime.date.today():
            return (False, ugettext('This coupon has expired.'))
        if self.numUses and self.allowedUses and self.allowedUses > 0 and self.numUses >= self.allowedUses:
            return (False, ugettext('This discount has exceeded the number of allowed uses.'))
            
        if cart:
            minOrder = self.minOrder or 0
            if cart.total < minOrder:
                return (False, ugettext('This discount only applies to orders of at least %s.' % moneyfmt(minOrder)))

            if not (self.allValid or (len(self._valid_products(cart.cartitem_set)) > 0)):
                return (False, ugettext('This discount cannot be applied to the products in your cart.'))
        
        # last minute check to make sure discount is valid
        success = {'valid' : True, 'message': ugettext('Valid.')}
        signals.discount_validate.send(
            sender=Discount, 
            discount=self, 
            cart=cart, 
            contact=contact, 
            success=success)
        return (success['valid'], success['message'])


    def _valid_products(self, item_query):
        validslugs = self.validProducts.values_list('slug', flat=True)
        itemslugs = item_query.values_list('product__slug', flat=True)
        return ProductPriceLookup.objects.filter(
            Q(discountable=True)
            &Q(productslug__in=validslugs)
            &Q(productslug__in=itemslugs)
            ).values_list('productslug', flat=True)

    def calc(self, order):
        # Use the order details and the discount specifics to calculate the actual discount
        discounted = {}
        if self.allValid:
            allvalid = True
        else:
            allvalid = False
            validproducts = self._valid_products(order.orderitem_set)

        for lineitem in order.orderitem_set.all():
            lid = lineitem.id
            price = lineitem.line_item_price
            if lineitem.product.is_discountable and (allvalid or lineitem.product.slug in validproducts):
                discounted[lid] = price
        signals.discount_filter_items.send(
            sender=self,
            discounted=discounted,
            order=order
            )

        if not self.shipping:
            self.shipping = "NONE"
            self.save()
            
        if self.shipping == "APPLY":
            shipcost = order.shipping_cost
            discounted['Shipping'] = shipcost

        if self.amount:
            # perform a flat rate discount, applying evenly to all items
            # in cart
            discounted = Discount.apply_even_split(discounted, self.amount)

        elif self.percentage:
            discounted = Discount.apply_percentage(discounted, self.percentage)

        else:
            # no discount, probably shipping-only
            zero = Decimal("0.00")
            for key in discounted.keys():
                discounted[key] = zero

        if self.shipping in ('FREE', 'FREECHEAP'):
            shipcost = order.shipping_cost
            discounted['Shipping'] = shipcost

        self._item_discounts = discounted
        self._calculated = True

    def save(self, **kwargs):
        if self.automatic:
            today = datetime.date.today()
            keyedcache.cache_delete('discount', 'sale', self.site, today)
        super(Discount, self).save(**kwargs)
        

    def _total(self):
        assert(self._calculated)
        return reduce(operator.add, self.item_discounts.values())
    total = property(_total)

    def _item_discounts(self):
        """Get the dictionary of orderitem -> discounts."""
        assert(self._calculated)
        return self._item_discounts
    item_discounts = property(_item_discounts)

    def _percentage_text(self):
        """Get the human readable form of the sale percentage."""
        return "%d%%" % self.percentage

    percentage_text = property(_percentage_text)

    def valid_for_product(self, product):
        """Tests if discount is valid for a single product"""
        if not product.is_discountable:
            return False
        p = self.validProducts.filter(id__exact = product.id)
        return p.count() > 0

    class Meta:
        verbose_name = _("Discount")
        verbose_name_plural = _("Discounts")

    def apply_even_split(cls, discounted, amount):
        lastct = -1
        ct = len(discounted)
        work = {}
        if ct > 0:
            split_discount = amount/ct
        else:
            split_discount = Decimal("0.00")

        while ct > 0:
            log.debug("Trying with ct=%i", ct)
            delta = Decimal("0.00")
            applied = Decimal("0.00")
            work = {}
            for lid, price in discounted.items():
                if price > split_discount:
                    work[lid] = split_discount
                    applied += split_discount
                else:
                    work[lid] = price
                    delta += price
                    applied += price
                    ct -= 1

            if applied >= amount - Decimal("0.01"):
                ct = 0

            if ct == lastct:
                ct = 0
            else:
                lastct = ct

            if ct > 0:
                split_discount = (amount-delta)/ct

        round_cents(work)
        return work

    apply_even_split = classmethod(apply_even_split)

    def apply_percentage(cls, discounted, percentage):
        work = {}

        for lid, price in discounted.items():
            work[lid] = price * percentage / 100
        round_cents(work)
        return work
        
    apply_percentage = classmethod(apply_percentage)


class OptionGroup(models.Model):
    """
    A set of options that can be applied to an item.
    Examples - Size, Color, Shape, etc
    """
    site = models.ForeignKey(Site, verbose_name=_('Site'))
    name = models.CharField(_("Name of Option Group"), max_length=50, 
        help_text=_("This will be the text displayed on the product page."))
    description = models.CharField(_("Detailed Description"), max_length=100,
        blank=True,
        help_text=_("Further description of this group (i.e. shirt size vs shoe size)."))
    sort_order = models.IntegerField(_("Sort Order"),
        help_text=_("The display order for this group."), default=0)

    objects = OptionGroupManager()

    def translated_description(self, language_code=None):
        return lookup_translation(self, 'description', language_code)

    def translated_name(self, language_code=None):
        return lookup_translation(self, 'name', language_code)

    def __unicode__(self):
        if self.description:
            return u"%s - %s" % (self.name, self.description)
        else:
            return self.name

    class Meta:
        ordering = ['sort_order', 'name']
        verbose_name = _("Option Group")
        verbose_name_plural = _("Option Groups")

class OptionGroupTranslation(models.Model):
    """A specific language translation for an `OptionGroup`.  This is intended for all descriptions which are not the
    default settings.LANGUAGE.
    """
    optiongroup = models.ForeignKey(OptionGroup, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    name = models.CharField(_("Translated OptionGroup Name"), max_length=255, )
    description = models.TextField(_("Description of OptionGroup"), default='', blank=True)
    version = models.IntegerField(_('version'), default=1)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('Option Group Translation')
        verbose_name_plural = _('Option Groups Translations')
        ordering = ('optiongroup', 'name','languagecode')
        unique_together = ('optiongroup', 'languagecode', 'version')

    def __unicode__(self):
        return u"OptionGroupTranslation: [%s] (ver #%i) %s Name: %s" % (self.languagecode, self.version, self.optiongroup, self.name)


class OptionManager(models.Manager):
    def from_unique_id(self, unique_id):
        group_id, option_value = split_option_unique_id(unique_id)
        group = OptionGroup.objects.get(id=group_id)
        return Option.objects.get(option_group=group_id, value=option_value)

class Option(models.Model):
    """
    These are the actual items in an OptionGroup.  If the OptionGroup is Size, then an Option
    would be Small.
    """
    objects = OptionManager()
    option_group = models.ForeignKey(OptionGroup)
    name = models.CharField(_("Display value"), max_length=50, )
    value = models.CharField(_("Stored value"), max_length=50)
    price_change = CurrencyField(_("Price Change"), null=True, blank=True,
        max_digits=14, decimal_places=6,
        help_text=_("This is the price differential for this option."))
    sort_order = models.IntegerField(_("Sort Order"), default=0)

    def translated_name(self, language_code=None):
        return lookup_translation(self, 'name', language_code)

    class Meta:
        ordering = ('option_group', 'sort_order', 'name')
        unique_together = (('option_group', 'value'),)
        verbose_name = _("Option Item")
        verbose_name_plural = _("Option Items")

    def _get_unique_id(self):
        return make_option_unique_id(self.option_group_id, self.value)
    unique_id = property(_get_unique_id)

    def __repr__(self):
        return u"<Option: %s>" % self.name

    def __unicode__(self):
        return u'%s: %s' % (self.option_group.name, self.name)

class OptionTranslation(models.Model):
    """A specific language translation for an `Option`.  This is intended for all descriptions which are not the
    default settings.LANGUAGE.
    """
    option = models.ForeignKey(Option, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    name = models.CharField(_("Translated Option Name"), max_length=255, )
    version = models.IntegerField(_('version'), default=1)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('Option Translation')
        verbose_name_plural = _('Option Translations')
        ordering = ('option', 'name','languagecode')
        unique_together = ('option', 'languagecode', 'version')

    def __unicode__(self):
        return u"OptionTranslation: [%s] (ver #%i) %s Name: %s" % (self.languagecode, self.version, self.option, self.name)

class ProductManager(models.Manager):
    
    def active(self, variations=True, **kwargs):
        if not variations:
            kwargs['productvariation__parent__isnull'] = True
        return self.filter(active=True, **kwargs)

    def active_by_site(self, variations=True, **kwargs):
        return self.by_site(active=True, variations=variations, **kwargs)

    def by_site(self, site=None, variations=True, **kwargs):
        if not site:
            site = Site.objects.get_current()
        
        site = site.id

        #log.debug("by_site: site=%s", site)
        if not variations:
            kwargs['productvariation__parent__isnull'] = True
        return self.filter(site__id__exact=site, **kwargs)

    def featured_by_site(self, site=None, **kwargs):
        return self.by_site(site=site, active=True, featured=True, **kwargs)

    def get_by_site(self, site=None, **kwargs):
        if not site:
            site = Site.objects.get_current()
        return self.get(site = site, **kwargs)

            
    def recent_by_site(self, **kwargs):
        query = self.active_by_site(**kwargs)
        if query.count() == 0:
            query = self.active_by_site()
            
        query = query.order_by('-date_added', '-id')
        return query
    

class Product(models.Model):
    """
    Root class for all Products
    """
    site = models.ForeignKey(Site, verbose_name=_('Site'))
    name = models.CharField(_("Full Name"), max_length=255, blank=False,
        help_text=_("This is what the product will be called in the default site language.  To add non-default translations, use the Product Translation section below."))
    slug = models.SlugField(_("Slug Name"), blank=True,
        help_text=_("Used for URLs, auto-generated from name if blank"), max_length=255)
    sku = models.CharField(_("SKU"), max_length=255, blank=True, null=True,
        help_text=_("Defaults to slug if left blank"))
    short_description = models.TextField(_("Short description of product"), help_text=_("This should be a 1 or 2 line description in the default site language for use in product listing screens"), max_length=200, default='', blank=True)
    description = models.TextField(_("Description of product"), help_text=_("This field can contain HTML and should be a few paragraphs in the default site language explaining the background of the product, and anything that would help the potential customer make their purchase."), default='', blank=True)
    category = models.ManyToManyField(Category, blank=True, verbose_name=_("Category"))
    items_in_stock = models.DecimalField(_("Number in stock"),  max_digits=18, decimal_places=6, default='0')
    meta = models.TextField(_("Meta Description"), max_length=200, blank=True, null=True, help_text=_("Meta description for this product"))
    date_added = models.DateField(_("Date added"), null=True, blank=True)
    active = models.BooleanField(_("Active"), default=True, help_text=_("This will determine whether or not this product will appear on the site"))
    featured = models.BooleanField(_("Featured"), default=False, help_text=_("Featured items will show on the front page"))
    ordering = models.IntegerField(_("Ordering"), default=0, help_text=_("Override alphabetical order in category display"))
    weight = models.DecimalField(_("Weight"), max_digits=8, decimal_places=2, null=True, blank=True)
    weight_units = models.CharField(_("Weight units"), max_length=3, null=True, blank=True) 
    #, validator_list=[weight_validator])
    length = models.DecimalField(_("Length"), max_digits=6, decimal_places=2, null=True, blank=True)
    length_units = models.CharField(_("Length units"), max_length=3, null=True, blank=True)
    #, validator_list=[length_validator])
    width = models.DecimalField(_("Width"), max_digits=6, decimal_places=2, null=True, blank=True)
    width_units = models.CharField(_("Width units"), max_length=3, null=True, blank=True)
    #, validator_list=[width_validator])
    height = models.DecimalField(_("Height"), max_digits=6, decimal_places=2, null=True, blank=True)
    height_units = models.CharField(_("Height units"), max_length=3, null=True, blank=True)
    #, validator_list=[height_validator])
    related_items = models.ManyToManyField('self', blank=True, null=True, verbose_name=_('Related Items'), related_name='related_products')
    also_purchased = models.ManyToManyField('self', blank=True, null=True, verbose_name=_('Previously Purchased'), related_name='also_products')
    total_sold = models.DecimalField(_("Total sold"),  max_digits=18, decimal_places=6, default='0')
    taxable = models.BooleanField(_("Taxable"), default=lambda: config_value('TAX', 'PRODUCTS_TAXABLE_BY_DEFAULT'))
    taxClass = models.ForeignKey('TaxClass', verbose_name=_('Tax Class'), blank=True, null=True, help_text=_("If it is taxable, what kind of tax?"))
    shipclass = models.CharField(_('Shipping'), choices=SHIP_CLASS_CHOICES, default="DEFAULT", max_length=10,
        help_text=_("If this is 'Default', then we'll use the product type to determine if it is shippable."))

    objects = ProductManager()

    def _get_mainCategory(self):
        """Return the first category for the product"""

        if self.category.count() > 0:
            return self.category.all()[0]
        else:
            return None

    main_category = property(_get_mainCategory)

    def _get_mainImage(self):
        img = False
        if self.productimage_set.count() > 0:
            img = self.productimage_set.order_by('sort')[0]
        else:
            # try to get a main image by looking at the parent if this has one
            p = self.get_subtype_with_attr('parent', 'product')
            if p:
                img = p.parent.product.main_image

        if not img:
            #This should be a "Image Not Found" placeholder image
            try:
                img = ProductImage.objects.filter(product__isnull=True).order_by('sort')[0]
            except IndexError:
                import sys
                print >>sys.stderr, 'Warning: default product image not found - try syncdb'

        return img

    main_image = property(_get_mainImage)
    
    def _is_discountable(self):
        p = self.get_subtype_with_attr('discountable')
        if p:
            return p.discountable
        else:
            return True
            
    is_discountable = property(_is_discountable)

    def translated_attributes(self, language_code=None):
        if not language_code:
            language_code = get_language()
        q = self.productattribute_set.filter(languagecode__exact = language_code)
        if q.count() == 0:
            q = self.productattribute_set.filter(Q(languagecode__isnull = True) | Q(languagecode__exact = ""))
        return q

    def translated_description(self, language_code=None):
        return lookup_translation(self, 'description', language_code)

    def translated_name(self, language_code=None):
        return lookup_translation(self, 'name', language_code)

    def translated_short_description(self, language_code=None):
        return lookup_translation(self, 'short_description', language_code)

    def _get_fullPrice(self):
        """
        returns price as a Decimal
        """

        subtype = self.get_subtype_with_attr('unit_price')

        if subtype and subtype is not self:
            price = subtype.unit_price
        else:
            price = get_product_quantity_price(self, Decimal('1'))

        if not price:
            price = Decimal("0.00")
        return price

    unit_price = property(_get_fullPrice)

    def get_qty_price(self, qty, include_discount=True):
        """
        If QTY_DISCOUNT prices are specified, then return the appropriate discount price for
        the specified qty.  Otherwise, return the unit_price
        returns price as a Decimal
        """
        subtype = self.get_subtype_with_attr('get_qty_price')
        if subtype and subtype is not self:
            price = subtype.get_qty_price(qty, include_discount=include_discount)

        else:
            if include_discount:
                price = get_product_quantity_price(self, qty)
            else:
                adjustment = get_product_quantity_adjustments(self, qty)
                if adjustment.price is not None:
                    price = adjustment.price.price
                else:
                    price = None
            if not price:
                price = self._get_fullPrice()

        return price
        
    def get_qty_price_list(self):
        """Return a list of tuples (qty, price)"""
        prices = Price.objects.filter(
            product__id=self.id).exclude(
            expires__isnull=False, 
            expires__lt=datetime.date.today()
        ).select_related()
        return [(price.quantity, price.dynamic_price) for price in prices]

    def in_stock(self):
        subtype = self.get_subtype_with_attr('in_stock')
        if subtype and subtype is not self:
            return subtype.in_stock

        return self.items_in_stock > 0

    def _has_full_dimensions(self):
        """Return true if the dimensions all have units and values. Used in shipping calcs. """
        for att in ('length', 'length_units', 'width', 'width_units', 'height', 'height_units'):
            if self.smart_attr(att) is None:
                return False
        return True

    has_full_dimensions = property(_has_full_dimensions)

    def _has_full_weight(self):
        """Return True if we have weight and weight units"""
        for att in ('weight', 'weight_units'):
            if self.smart_attr(att) is None:
                return False
        return True

    has_full_weight = property(_has_full_weight)

    def __unicode__(self):
        return self.name

    def get_absolute_url(self):
        return urlresolvers.reverse('satchmo_product', kwargs={'product_slug': self.slug})

    class Meta:
        ordering = ('site', 'ordering', 'name')
        verbose_name = _("Product")
        verbose_name_plural = _("Products")
        unique_together = (('site', 'sku'),('site','slug'))

    def save(self, **kwargs):
        if not self.pk:
            self.date_added = datetime.date.today()

        if self.name and not self.slug:
            self.slug = slugify(self.name, instance=self)

        if not self.sku:
            self.sku = self.slug
        super(Product, self).save(**kwargs)
        ProductPriceLookup.objects.smart_create_for_product(self)

    def get_subtypes(self):
        # If we've already computed it once, let's not do it again.
        # This is a performance speedup.
        if hasattr(self,"_sub_types"):
            return self._sub_types
        types = []
        try:
            for key in config_value('PRODUCT', 'PRODUCT_TYPES'):
                app, subtype = key.split("::")
                try:
                    subclass = getattr(self, subtype.lower())
                    gettype = getattr(subclass, '_get_subtype')
                    subtype = gettype()
                    if not subtype in types:
                        types.append(subtype)
                except models.ObjectDoesNotExist:
                    pass
        except SettingNotSet:
            log.warn("Error getting subtypes, OK if in SyncDB")

        self._sub_types = tuple(types)
        return self._sub_types

    get_subtypes.short_description = _("Product Subtypes")

    def get_subtype_with_attr(self, *args):
        """Get a subtype with the specified attributes.  Note that this can be chained
        so that you can ensure that the attribute then must have the specified attributes itself.

        example:  get_subtype_with_attr('parent') = any parent
        example:  get_subtype_with_attr('parent', 'product') = any parent which has a product attribute
        """
        for subtype_name in self.get_subtypes():
            subtype = getattr(self, subtype_name.lower())
            if hasattr(subtype, args[0]):
                if len(args) == 1:
                    return subtype
                else:
                    found = True
                    for attr in args[1:-1]:
                        if hasattr(subtype, attr):
                            subtype = getattr(self, attr)
                        else:
                            found = False
                            break
                    if found and hasattr(subtype, args[-1]):
                        return subtype

        return None

    def smart_attr(self, attr):
        """Retrieve an attribute, or its parent's attribute if it is null or blank.
        Ex: to get a weight.  obj.smart_attr('weight')"""

        val = getattr(self, attr)
        if val is None or val == "":
            for subtype_name in self.get_subtypes():
                subtype = getattr(self, subtype_name.lower())

                if hasattr(subtype, 'parent'):
                    subtype = subtype.parent.product

                if hasattr(subtype, attr):
                    val = getattr(subtype, attr)
                    if val is not None:
                        break
        return val
                        
    def smart_relation(self, relation):
        """Retrieve a relation, or its parent's relation if the relation count is 0"""
        q = getattr(self, relation)
        if q.count() > 0:
            return q
        
        for subtype_name in self.get_subtypes():
            subtype = getattr(self, subtype_name.lower())

            if hasattr(subtype, 'parent'):
                subtype = subtype.parent.product

                return getattr(subtype, relation)

    def _has_variants(self):
        subtype = self.get_subtype_with_attr('has_variants')
        return subtype and subtype.has_variants

    has_variants = property(_has_variants)

    def _get_category(self):
        """
        Return the primary category associated with this product
        """
        subtype = self.get_subtype_with_attr('get_category')
        if subtype and subtype is not self:
            return subtype.get_category

        try:
            return self.category.all()[0]
        except IndexError:
            return None

    get_category = property(_get_category)

    def _get_downloadable(self):
        """
        If this Product has any subtypes associated with it that are downloadable, then
        consider it downloadable
        """
        return self.get_subtype_with_attr('is_downloadable') is not None

    is_downloadable = property(_get_downloadable)

    def _get_subscription(self):
        """
        If this Product has any subtypes associated with it that are subscriptions, then
        consider it subscription based.
        """
        for prod_type in self.get_subtypes():
            subtype = getattr(self, prod_type.lower())
            if hasattr(subtype, 'is_subscription'):
                return subtype.is_subscription
        return False
    is_subscription = property(_get_subscription)

    def _get_shippable(self):
        """
        If this Product has any subtypes associated with it that are not
        shippable, then consider the product not shippable.
        If it is downloadable, then we don't ship it either.
        """
        if self.shipclass=="DEFAULT":
            subtype = self.get_subtype_with_attr('is_shippable')
            if subtype and subtype is not self and not subtype.is_shippable:
                return False
            return True
        elif self.shipclass=="YES":
            return True
        else:
            return False
            
    is_shippable = property(_get_shippable)

    def add_template_context(self, context, *args, **kwargs):
        """
        Add context for the product template.
        Call the add_template_context method of each subtype and return the
        combined context.
        """
        subtypes = self.get_subtypes()
        logging.debug('subtypes = %s', subtypes)
        for subtype_name in subtypes:
            subtype = getattr(self, subtype_name.lower())
            if subtype == self:
                continue
            if hasattr(subtype, 'add_template_context'):
                context = subtype.add_template_context(context, *args, **kwargs)

        return context

class ProductTranslation(models.Model):
    """A specific language translation for a `Product`.  This is intended for all descriptions which are not the
    default settings.LANGUAGE.
    """
    product = models.ForeignKey('Product', related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    name = models.CharField(_("Full Name"), max_length=255)
    description = models.TextField(_("Description of product"), help_text=_("This field can contain HTML and should be a few paragraphs explaining the background of the product, and anything that would help the potential customer make their purchase."), default='', blank=True)
    short_description = models.TextField(_("Short description of product"), help_text=_("This should be a 1 or 2 line description for use in product listing screens"), max_length=200, default='', blank=True)
    version = models.IntegerField(_('version'), default=1)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('Product Translation')
        verbose_name_plural = _('Product Translations')
        ordering = ('product', 'name','languagecode')
        unique_together = ('product', 'languagecode', 'version')

    def __unicode__(self):
        return u"ProductTranslation: [%s] (ver #%i) %s Name: %s" % (self.languagecode, self.version, self.product, self.name)

def get_all_options(obj, ids_only=False):
    """
    Returns all possible combinations of options for this products OptionGroups as a List of Lists.
    Ex:
    For OptionGroups Color and Size with Options (Blue, Green) and (Large, Small) you'll get
    [['Blue', 'Small'], ['Blue', 'Large'], ['Green', 'Small'], ['Green', 'Large']]
    Note: the actual values will be instances of Option instead of strings
    """
    sublist = []
    masterlist = []
    #Create a list of all the options & create all combos of the options
    for opt in obj.option_group.select_related().all():
        for value in opt.option_set.all():
            if ids_only:
                sublist.append(value.unique_id)
            else:
                sublist.append(value)
        masterlist.append(sublist)
        sublist = []
    results = cross_list(masterlist)
    return results

class CustomProduct(models.Model):
    """
    Product which must be custom-made or ordered.
    """
    product = models.OneToOneField(Product, verbose_name=_('Product'), primary_key=True)
    downpayment = models.IntegerField(_("Percent Downpayment"), default=20)
    deferred_shipping = models.BooleanField(_('Deferred Shipping'),
        help_text=_('Do not charge shipping at checkout for this item.'),
        default=False)
    option_group = models.ManyToManyField(OptionGroup, verbose_name=_('Option Group'), blank=True,)

    def _is_shippable(self):
        return not self.deferred_shipping
    is_shippable = property(fget=_is_shippable)

    def _get_fullPrice(self):
        """
        returns price as a Decimal
        """
        return self.get_qty_price(Decimal('1'))

    unit_price = property(_get_fullPrice)

    def add_template_context(self, context, selected_options, **kwargs):
        """
        Add context for the product template.
        Return the updated context.
        """
        from product.utils import serialize_options

        context['options'] = serialize_options(self, selected_options)

        return context

    def get_qty_price(self, qty, include_discount=True):
        """
        If QTY_DISCOUNT prices are specified, then return the appropriate discount price for
        the specified qty.  Otherwise, return the unit_price
        returns price as a Decimal
        """
        if include_discount:
            price = get_product_quantity_price(self.product, qty)
        else:
            adjustment = get_product_quantity_adjustments(self, qty)
            if adjustment.price is not None:
                price = adjustment.price.price
            else:
                price = None

        if not price and qty == Decimal('1'): # Prevent a recursive loop.
            price = Decimal("0.00")
        elif not price:      
            price = self.product._get_fullPrice()

        return price * self.downpayment / 100

    def get_full_price(self, qty=Decimal('1')):
        """
        Return the full price, ignoring the deposit.
        """
        price = get_product_quantity_price(self.product, qty)
        if not price:
            price = self.product.unit_price

        return price

    full_price = property(fget=get_full_price)



    def _get_subtype(self):
        return 'CustomProduct'

    def __unicode__(self):
        return u"CustomProduct: %s" % self.product.name

    def get_valid_options(self):
        """
        Returns all of the valid options
        """
        return get_all_options(self, ids_only=True)
        
    def save(self):
        if hasattr(self.product,'_sub_types'):
            del self.product._sub_types
        super(CustomProduct, self).save()


    class Meta:
        verbose_name = _('Custom Product')
        verbose_name_plural = _('Custom Products')

class CustomTextField(models.Model):
    """
    A text field to be filled in by a customer.
    """

    name = models.CharField(_('Custom field name'), max_length=40, )
    slug = models.SlugField(_("Slug"), help_text=_("Auto-generated from name if blank"),
        blank=True)
    products = models.ForeignKey(CustomProduct, verbose_name=_('Custom Fields'), 
        related_name='custom_text_fields')
    sort_order = models.IntegerField(_("Sort Order"),
        help_text=_("The display order for this group."), default=0)
    price_change = CurrencyField(_("Price Change"), max_digits=14, 
        decimal_places=6, blank=True, null=True)

    def save(self, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name, instance=self)
        super(CustomTextField, self).save(**kwargs)

    def translated_name(self, language_code=None):
        return lookup_translation(self, 'name', language_code)

    class Meta:
        ordering = ('sort_order',)

class CustomTextFieldTranslation(models.Model):
    """A specific language translation for a `CustomTextField`.  This is intended for all descriptions which are not the
    default settings.LANGUAGE.
    """
    customtextfield = models.ForeignKey(CustomTextField, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    name = models.CharField(_("Translated Custom Text Field Name"), max_length=255, )
    version = models.IntegerField(_('version'), default=1)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('CustomTextField Translation')
        verbose_name_plural = _('CustomTextField Translations')
        ordering = ('customtextfield', 'name','languagecode')
        unique_together = ('customtextfield', 'languagecode', 'version')

    def __unicode__(self):
        return u"CustomTextFieldTranslation: [%s] (ver #%i) %s Name: %s" % (self.languagecode, self.version, self.customtextfield, self.name)

class ConfigurableProduct(models.Model):
    """
    Product with selectable options.
    This is a sort of virtual product that is visible to the customer, but isn't actually stocked on a shelf,
    the specific "shelf" product is determined by the selected options.
    """
    product = models.OneToOneField(Product, verbose_name=_("Product"), primary_key=True)
    option_group = models.ManyToManyField(OptionGroup, blank=True, verbose_name=_("Option Group"))
    create_subs = models.BooleanField(_("Create Variations"), default=False, help_text =_("Create ProductVariations for all this product's options.  To use this, you must first add an option, save, then return to this page and select this option."))

    def __init__(self, *args, **kwargs):
        super(ConfigurableProduct, self).__init__(*args, **kwargs)

    def _get_subtype(self):
        return 'ConfigurableProduct'

    def get_all_options(self):
        """
        Returns all possible combinations of options for this products OptionGroups as a List of Lists.
        Ex:
        For OptionGroups Color and Size with Options (Blue, Green) and (Large, Small) you'll get
        [['Blue', 'Small'], ['Blue', 'Large'], ['Green', 'Small'], ['Green', 'Large']]
        Note: the actual values will be instances of Option instead of strings
        """
        sublist = []
        masterlist = []
        #Create a list of all the options & create all combos of the options
        for opt in self.option_group.all():
            for value in opt.option_set.all():
                sublist.append(value)
            masterlist.append(sublist)
            sublist = []
        return cross_list(masterlist)

    def get_valid_options(self):
        """
        Returns unique_ids from get_all_options(), but filters out Options that this
        ConfigurableProduct doesn't have a ProductVariation for.
        """
        variations = self.productvariation_set.filter(product__active='1')
        active_options = [v.unique_option_ids for v in variations]
        all_options = get_all_options(self, ids_only=True)
        return [opt for opt in all_options if self._unique_ids_from_options(opt) in active_options]

    def create_all_variations(self):
        """
        Get a list of all the optiongroups applied to this object
        Create all combinations of the options and create variations
        """
        # Create a new ProductVariation for each combination.
        for options in self.get_all_options():
            self.create_variation(options)

    def create_variation(self, options, name=u"", sku=u"", slug=u""):
        """Create a productvariation with the specified options.
        Will not create a duplicate."""
        log.debug("Create variation: %s", options)
        variations = self.get_variations_for_options(options)

        if not variations:
            # There isn't an existing ProductVariation.
            if self.product:
                site = self.product.site
            else:
                site = self.site
                
            variant = Product(site=site, items_in_stock=0, name=name)
            optnames = [opt.value for opt in options]
            if not slug:
                slug = slugify(u'%s_%s' % (self.product.slug, u'_'.join(optnames)))

            while Product.objects.filter(slug=slug).count():
                slug = u'_'.join((slug, unicode(self.product.id)))

            variant.slug = slug

            log.info("Creating variation for [%s] %s", self.product.slug, variant.slug)
            variant.save()

            pv = ProductVariation(product=variant, parent=self)
            pv.save()

            for option in options:
                pv.options.add(option)

            pv.name = name
            pv.sku = sku
            pv.save()

        else:
            variant = variations[0].product
            log.debug("Existing variant: %s", variant)
            dirty = False
            if name and name != variant.name:
                log.debug("Updating name: %s --> %s", self, name)
                variant.name = name
                dirty = True
            if sku and sku != variant.sku:
                variant.sku = sku
                log.debug("Updating sku: %s --> %s", self, sku)
                dirty = True
            if slug:
                # just in case
                slug = slugify(slug)
            if slug and slug != variant.slug:
                variant.slug = slug
                log.debug("Updating slug: %s --> %s", self, slug)
                dirty = True
            if dirty:
                log.debug("Changed existing variant, saving: %s", variant)
                variant.save()
            else:
                log.debug("No change to variant, skipping save: %s", variant)

        return variant

    def _unique_ids_from_options(self, options):
        """
        Takes an iterable of Options (or str(Option)) and outputs a sorted tuple of
        option unique ids suitable for comparing to a productvariation.option_values
        """
        optionlist = []
        for opt in options:
            if isinstance(options[0], Option):
                opt = opt.unique_id
            optionlist.append(opt)
            
        return sorted_tuple(optionlist)

    def get_product_from_options(self, options):
        """
        Accepts an iterable of either Option object or a sorted tuple of
        options ids.
        Returns the product that matches or None
        """    
        options = self._unique_ids_from_options(options)
        pv = None
        if hasattr(self, '_variation_cache'):
            pv =  self._variation_cache.get(options, None)
        else:
            for member in self.productvariation_set.all():
                if member.unique_option_ids == options:
                    pv = member
                    break
        if pv:
            return pv.product        
        return None

    def get_variations_for_options(self, options):
        """
        Returns a list of existing ProductVariations with the specified options.
        """
        variations = ProductVariation.objects.filter(parent=self)
        for option in options:
            variations = variations.filter(options=option)
        return variations

    def add_template_context(self, context, request, selected_options, default_view_tax=False, **kwargs):
        """
        Add context for the product template.
        Return the updated context.
        """
        from product.utils import productvariation_details, serialize_options

        selected_options = self._unique_ids_from_options(selected_options)
        context['options'] = serialize_options(self, selected_options)
        context['details'] = productvariation_details(self.product, default_view_tax,
            request.user)
                    
        return context

    def save(self, **kwargs):
        """
        Right now this only works if you save the suboptions, then go back and choose to create the variations.
        """
        super(ConfigurableProduct, self).save()
        if hasattr(self.product,'_sub_types'):
            del self.product._sub_types
        # Doesn't work with admin - the manipulator doesn't add the option_group
        # until after save() is called.
        if self.create_subs and self.option_group.count():
            self.create_all_variations()
            self.create_subs = False
            super(ConfigurableProduct, self).save(**kwargs)
            
        ProductPriceLookup.objects.smart_create_for_product(self.product)

    def get_absolute_url(self):
        return self.product.get_absolute_url()
        
    def setup_variation_cache(self):
        self._variation_cache = {}
        for member in self.productvariation_set.all():        
            key = member.unique_option_ids
            self._variation_cache[key] = member

    class Meta:
        verbose_name = _("Configurable Product")
        verbose_name_plural = _("Configurable Products")

    def __unicode__(self):
        return self.product.slug

def _protected_dir(instance, filename):
    raw = config_value_safe('PRODUCT', 'PROTECTED_DIR', 'images/')
    updir = normalize_dir(raw)
    return os.path.normpath(os.path.join(updir, os.path.basename(filename)))

class DownloadableProduct(models.Model):
    """
    This type of Product is a file to be downloaded
    """
    product = models.OneToOneField(Product, verbose_name=_("Product"), primary_key=True)
    file = FileField(_("File"), upload_to=_protected_dir)
    num_allowed_downloads = models.IntegerField(_("Num allowed downloads"), help_text=_("Number of times link can be accessed."))
    expire_minutes = models.IntegerField(_("Expire minutes"), help_text=_("Number of minutes the link should remain active."))
    active = models.BooleanField(_("Active"), help_text=_("Is this download currently active?"), default=True)
    is_shippable = False
    is_downloadable = True

    def __unicode__(self):
        return self.product.slug

    def _get_subtype(self):
        return 'DownloadableProduct'

    def create_key(self):
        salt = sha1(str(random.random())).hexdigest()[:5]
        download_key = sha1(salt+smart_str(self.product.name)).hexdigest()
        return download_key

    def order_success(self, order, order_item):
        signals.subtype_order_success.send(self, product=self, order=order, subtype="download")

    class Meta:
        verbose_name = _("Downloadable Product")
        verbose_name_plural = _("Downloadable Products")
    
    def save(self):
        if hasattr(self.product,'_sub_types'):
            del self.product._sub_types
        super(DownloadableProduct, self).save()

class SubscriptionProduct(models.Model):
    """
    This type of Product is for recurring billing (memberships, subscriptions, payment terms)
    """
    product = models.OneToOneField(Product, verbose_name=_("Product"), primary_key=True)
    recurring = models.BooleanField(_("Recurring Billing"), help_text=_("Customer will be charged the regular product price on a periodic basis."), default=False)
    recurring_times = models.IntegerField(_("Recurring Times"), help_text=_("Number of payments which will occur at the regular rate.  (optional)"), null=True, blank=True)
    expire_length = models.IntegerField(_("Duration"), help_text=_("Length of each billing cycle"), null=True, blank=True)
    SUBSCRIPTION_UNITS = (
        ('DAY', _('Days')),
        ('MONTH', _('Months'))
    )
    expire_unit = models.CharField(_("Expire Unit"), max_length=5, choices=SUBSCRIPTION_UNITS, default="DAY", null=False)
    SHIPPING_CHOICES = (
        ('0', _('No Shipping Charges')),
        ('1', _('Pay Shipping Once')),
        ('2', _('Pay Shipping Each Billing Cycle')),
    )
    is_shippable = models.IntegerField(_("Shippable?"), help_text=_("Is this product shippable?"), max_length=1, choices=SHIPPING_CHOICES)

    is_subscription = True

    def _get_subtype(self):
        return 'SubscriptionProduct'

    def __unicode__(self):
        return self.product.slug

    def _get_fullPrice(self):
        """
        returns price as a Decimal
        """
        return self.get_qty_price(1)

    unit_price = property(_get_fullPrice)

    def get_qty_price(self, qty, show_trial=True, include_discount=True):
        """
        If QTY_DISCOUNT prices are specified, then return the appropriate discount price for
        the specified qty.  Otherwise, return the unit_price
        returns price as a Decimal
        
        Note: If a subscription has a trial, then we'll return the first trial price, otherwise the checkout won't
        balance and it will look like there are items to be paid on the order.
        """
        if show_trial:
            trial = self.get_trial_terms(0)
        else:
            trial = None
            
        if trial:
            price = trial.price * qty
        else:
            if include_discount:
                price = get_product_quantity_price(self.product, qty)
            else:
                adjustment = get_product_quantity_adjustments(self, qty)
                if adjustment.price is not None:
                    price = adjustment.price.price
                else:
                    price = None

            if not price and qty == Decimal('1'):      # Prevent a recursive loop.
                price = Decimal("0.00")
            elif not price:      
                price = self.product._get_fullPrice()
        return price

    def recurring_price(self):
        """
        Get the non-trial price.
        """
        return self.get_qty_price(Decimal('1'), show_trial=False)

    # use order_success() and DownloadableProduct.create_key() to add user to group and perform other tasks
    def get_trial_terms(self, trial=None):
        """Get the trial terms for this subscription"""
        if trial is None:
            return self.trial_set.all().order_by('id')
        else:
            try:
                return self.trial_set.all().order_by('id')[trial]
            except IndexError:
                return None
                
    def calc_expire_date(self, date=None):
        if date is None:
            date = datetime.datetime.now()
        if self.expire_unit == "DAY":
            expiredate = date + datetime.timedelta(days=self.expire_length)
        else:
            expiredate = add_month(date, n=self.expire_length)
        
        return expiredate
    
    def save(self):
        if hasattr(self.product,'_sub_types'):
            del self.product._sub_types
        super(SubscriptionProduct, self).save()

    class Meta:
        verbose_name = _("Subscription Product")
        verbose_name_plural = _("Subscription Products")

class Trial(models.Model):
    """
    Trial billing terms for subscription products.
    Separating it out lets us have as many trial periods as we want.
    Note that some third party payment processors support only a limited number of trial
    billing periods.  For example, PayPal limits us to 2 trial periods, so if you are using
    PayPal for a billing option, you need to create no more than 2 trial periods for your
    product.  However, gateway based processors like Authorize.net can support as many
    billing periods as you wish.
    """
    subscription = models.ForeignKey(SubscriptionProduct)
    price = CurrencyField(_("Price"), help_text=_("Set to 0 for a free trial.  Leave empty if product does not have a trial."), max_digits=10, decimal_places=2, null=True, )
    expire_length = models.IntegerField(_("Trial Duration"), help_text=_("Length of trial billing cycle.  Leave empty if product does not have a trial."), null=True, blank=True)

    def __unicode__(self):
        return unicode(self.price)
        
    def _occurrences(self):
        if self.expire_length:
            return int(self.expire_length/self.subscription.expire_length)
        else:
            return 0
    occurrences = property(fget=_occurrences)

    def calc_expire_date(self, date=None):
        if date is None:
            date = datetime.datetime.now()
        if self.subscription.expire_unit == "DAY":
            expiredate = date + datetime.timedelta(days=self.expire_length)
        else:
            expiredate = add_month(date, n=self.expire_length)
        
        return expiredate

    class Meta:
        ordering = ['-id']
        verbose_name = _("Trial Terms")
        verbose_name_plural = _("Trial Terms")

#class BundledProduct(models.Model):
#    """
#    This type of Product is a group of products that are sold as a set
#    NOTE: This doesn't do anything yet - it's just an example
#    """
#    product = models.OneToOneField(Product)
#    members = models.ManyToManyField(Product, related_name='parent_productgroup_set')
#

class ProductVariationManager(models.Manager):

    def by_parent(self, parent):
        """Get the list of productvariations which have the `product` as the parent"""
        return ProductVariation.objects.filter(parent=parent)

class ProductVariation(models.Model):
    """
    This is the real Product that is ordered when a customer orders a
    ConfigurableProduct with the matching Options selected

    """
    product = models.OneToOneField(Product, verbose_name=_('Product'), primary_key=True)
    options = models.ManyToManyField(Option, verbose_name=_('Options'))
    parent = models.ForeignKey(ConfigurableProduct, 
    #, validator_list=[variant_validator]
    verbose_name=_('Parent'))

    objects = ProductVariationManager()

    def _get_fullPrice(self):
        """ Get price based on parent ConfigurableProduct """
        # allow explicit setting of prices.
        #qty_discounts = self.price_set.exclude(expires__isnull=False, expires__lt=datetime.date.today()).filter(quantity__lte=1)
        try:
            qty_discounts = Price.objects.filter(product__id=self.product.id).exclude(expires__isnull=False, expires__lt=datetime.date.today())
            if qty_discounts.count() > 0:
                # Get the price with the quantity closest to the one specified without going over
                return qty_discounts.order_by('-quantity')[0].dynamic_price

            if self.parent.product.unit_price is None:
                log.warn("%s: Unexpectedly no parent.product.unit_price", self)
                return None

        except AttributeError:
            pass

        # calculate from options
        return self.parent.product.unit_price + self.price_delta()

    unit_price = property(_get_fullPrice)

    def _get_optionName(self):
        "Returns the options in a human readable form"
        if self.options.count() == 0:
            return self.parent.verbose_name
        output = self.parent.verbose_name + " ( "
        numProcessed = 0
        # We want the options to be sorted in a consistent manner
        optionDict = dict([(sub.option_group.sort_order, sub) for sub in self.options.all()])
        for optionNum in optionDict.keys().sort():
            numProcessed += 1
            if numProcessed == self.options.count():
                output += optionDict[optionNum].name
            else:
                output += optionDict[optionNum].name + "/"
        output += " )"
        return output
    full_name = property(_get_optionName)

    def _optionkey(self):
        #todo: verify ordering
        optkeys = [str(x) for x in self.options.values_list('value', flat=True).order_by('option_group__id')]
        return "::".join(optkeys)
    optionkey = property(fget=_optionkey)

    def _get_option_ids(self):
        """
        Return a sorted tuple of all the valid options for this variant.
        """
        qry = self.options.values_list('option_group__id', 'value').order_by('option_group')
        ret = [make_option_unique_id(*v) for v in qry]
        return sorted_tuple(ret)
        
    unique_option_ids = property(_get_option_ids)

    def _get_subtype(self):
        return 'ProductVariation'

    def _has_variants(self):
        return True
    has_variants = property(_has_variants)

    def _get_category(self):
        """
        Return the primary category associated with this product
        """
        return self.parent.product.category.all()[0]
    get_category = property(_get_category)

    def _check_optionParents(self):
        groupList = []
        for option in self.options.all():
            if option.option_group.id in groupList:
                return(True)
            else:
                groupList.append(option.option_group.id)
        return(False)

    def get_qty_price(self, qty, include_discount=True):
        if include_discount:
            price = get_product_quantity_price(
                self.product, qty, 
                delta=self.price_delta(False), 
                parent=self.parent.product)
        else:
            adjustment = get_product_quantity_adjustments(self, qty, parent=self.parent.product)
            if adjustment.price is not None:
                price = adjustment.price.price + self.price_delta(True)
            else:
                price = None

        return price

    def get_qty_price_list(self):
        """Return a list of tuples (qty, price)"""
        prices = Price.objects.filter(product__id=self.product.id).exclude(expires__isnull=False, expires__lt=datetime.date.today())
        if prices.count() > 0:
            # prices directly set, return them
            pricelist = [(price.quantity, price.dynamic_price) for price in prices]
        else:
            prices = self.parent.product.get_qty_price_list()
            price_delta = self.price_delta()

            pricelist = [(qty, price+price_delta) for qty, price in prices]

        return pricelist
        
    def _is_shippable(self):
        product = self.product
        parent = self.parent.product
        return ((product.shipclass == "DEFAULT" and parent.shipclass == "DEFAULT")
                or product.shipclass == 'YES')

    is_shippable = property(fget=_is_shippable)
    
    def isValidOption(self, field_data, all_data):
        raise forms.ValidationError(_("Two options from the same option group cannot be applied to an item."))

    def price_delta(self, include_discount=True):
        # TODO: deltas aren't taken into account by satchmo_price_query
        price_delta = Decimal("0.00")
        for option in self.options.all():
            if option.price_change:
                price_delta += Decimal(option.price_change)
        return price_delta

    def save(self, **kwargs):
        # don't save if the product is a configurableproduct
        if "ConfigurableProduct" in self.product.get_subtypes():
            log.warn("cannot add a productvariation subtype to a product which already is a configurableproduct. Aborting")
            return
            
        pvs = ProductVariation.objects.filter(parent=self.parent)
        pvs = pvs.exclude(product=self.product)
        for pv in pvs:
            if pv.unique_option_ids == self.unique_option_ids:
                return # Don't allow duplicates

        if not self.product.name:
            # will force calculation of default name
            self.name = ""

        super(ProductVariation, self).save(**kwargs)
        ProductPriceLookup.objects.smart_create_for_product(self.product)

    def _set_name(self, name):
        if not name:
            name = self.parent.product.name
            options = [option.name for option in self.options.order_by("option_group")]
            if options:
                name = u'%s (%s)' % (name, u'/'.join(options))
            log.debug("Setting default name for ProductVariant: %s", name)

        self.product.name = name
        self.product.save()

    def _get_name(self):
        return self.product.name

    name = property(fset=_set_name, fget=_get_name)

    def _set_sku(self, sku):
        if not sku:
            sku = self.product.slug
        self.product.sku = sku
        self.product.save()

    def _get_sku(self):
        return self.product.sku

    sku = property(fset=_set_sku, fget=_get_sku)

    def get_absolute_url(self):
        return self.product.get_absolute_url()

    class Meta:
        verbose_name = _("Product variation")
        verbose_name_plural = _("Product variations")

    def __unicode__(self):
        return self.product.slug
        

class ProductPriceLookupManager(models.Manager):
    
    def by_product(self, product):
        return self.get(productslug=product.slug)
    
    def delete_expired(self):
        for p in self.filter(expires__lt=datetime.date.today()):
            p.delete()
    
    def create_for_product(self, product):
        """Create a set of lookup objects for all priced quantities of the Product"""

        self.delete_for_product(product)
        pricelist = product.get_qty_price_list()
            
        objs = []
        for qty, price in pricelist:
            obj = ProductPriceLookup(productslug=product.slug, 
                siteid=product.site_id,
                active=product.active,
                price=price, 
                quantity=qty, 
                discountable=product.is_discountable,
                items_in_stock=product.items_in_stock)
            obj.save()
            objs.append(obj)
        return objs
        
    def create_for_configurableproduct(self, configproduct):
        """Create a set of lookup objects for all variations of this product"""

        objs = self.create_for_product(configproduct)
        for pv in configproduct.configurableproduct.productvariation_set.filter(product__active='1'):
            objs.extend(self.create_for_variation(pv, configproduct))
        
        return objs
        
    def create_for_variation(self, variation, parent):
        
        product = variation.product
        
        self.delete_for_product(product)
        pricelist = variation.get_qty_price_list()
            
        objs = []
        for qty, price in pricelist:
            obj = ProductPriceLookup(productslug=product.slug, 
                parentid=parent.pk,
                siteid=product.site_id,
                active=product.active,
                price=price, 
                quantity=qty, 
                key=variation.optionkey, 
                discountable=product.is_discountable,
                items_in_stock=product.items_in_stock)
            obj.save()
            objs.append(obj)
        return objs
        
    def delete_for_product(self, product):
        for obj in self.filter(productslug=product.slug):
            obj.delete()
        
    def rebuild_all(self, site=None):
        if not site:
            site = Site.objects.get_current()

        for lookup in self.filter(siteid=site.id):
            lookup.delete()

        ct = 0
        log.debug('ProductPriceLookup rebuilding all pricing')
        for p in Product.objects.active_by_site(site=site, variations=False):
            prices = self.smart_create_for_product(p)
            ct += len(prices)
        log.info('ProductPriceLookup built %i prices', ct)
            
    def smart_create_for_product(self, product):
        subtypes = product.get_subtypes()
        if 'ConfigurableProduct' in subtypes:
            return self.create_for_configurableproduct(product)
        elif 'ProductVariation' in subtypes:
            return self.create_for_variation(product.productvariation, product.productvariation.parent)
        else:
            return self.create_for_product(product)
    
class ProductPriceLookup(models.Model):
    """
    A denormalized object, used to quickly provide
    details needed for productvariation display, without way too many database hits.
    """
    siteid = models.IntegerField()
    key = models.CharField(max_length=60, null=True)
    parentid = models.IntegerField(null=True)
    productslug = models.CharField(max_length=255)
    price = models.DecimalField(max_digits=14, decimal_places=6)
    quantity = models.DecimalField(max_digits=18, decimal_places=6)
    active = models.BooleanField()
    discountable = models.BooleanField()
    items_in_stock = models.DecimalField(max_digits=18, decimal_places=6)

    objects = ProductPriceLookupManager()
    
    def _product(self):
        return Product.objects.get(slug=self.productslug)
        
    product = property(fget=_product)    
    
    def _dynamic_price(self):
        """Get the current price as modified by all listeners."""
        adjust = PriceAdjustmentCalc(self)
        signals.satchmo_price_query.send(self, adjustment=adjust, 
            slug=self.productslug, discountable=self.discountable)
        self.price = adjust.final_price()
        return self.price

    dynamic_price = property(fget=_dynamic_price)

# Support the user's setting of custom expressions in the settings.py file
try:
    user_validations = settings.SATCHMO_SETTINGS.get('ATTRIBUTE_VALIDATIONS')
except:
    user_validations = None
            
VALIDATIONS = [
            ('product.utils.validation_simple', _('One or more characters')),
            ('product.utils.validation_integer', _('Integer number')),
            ('product.utils.validation_yesno', _('Yes or No')),
            ('product.utils.validation_decimal', _('Decimal number')),
            ]
if user_validations: 
    VALIDATIONS.extend(user_validations)

class AttributeOption(models.Model):
    """
    Allows arbitrary name/value pairs to be attached to a product.
    By defining the list, the user will be presented with a predefined
    list of attributes instead of a free form field.
    The validation field should contain a regular expression that can be 
    used to validate the structure of the input.
    Possible usage for a book:
    ISBN, Pages, Author, etc
    """
    description = models.CharField(_("Description"), max_length=100)
    name = models.SlugField(_("Attribute name"), max_length=100)
    validation = models.CharField(_("Field Validations"), choices=VALIDATIONS, max_length=100)
    sort_order = models.IntegerField(_("Sort Order"), default=1)
    error_message = models.CharField(_("Error Message"), default=_("Invalid Entry"), max_length=100)
    
    class Meta:
        ordering = ('sort_order',)
        
    def __unicode__(self):
        return self.description
    

class ProductAttribute(models.Model):
    """
    Allows arbitrary name/value pairs (as strings) to be attached to a product.
    This is a simple way to add extra text or numeric info to a product.
    If you want more structure than this, create your own subtype to add
    whatever you want to your Products.
    """
    product = models.ForeignKey(Product)
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES, null=True, blank=True)
    option = models.ForeignKey(AttributeOption, null=True, blank=True)
    value = models.CharField(_("Value"), max_length=255, null=True, blank=True)

    def _name(self):
        return self.option.name
    name = property(_name)
    
    def _description(self):
        return self.option.description
    description = property(_description)
    
    class Meta:
        verbose_name = _("Product Attribute")
        verbose_name_plural = _("Product Attributes")
        
    def __unicode__(self):
        return self.option.name

class Price(models.Model):
    """
    A Price!
    Separating it out lets us have different prices for the same product for different purposes.
    For example for quantity discounts.
    The current price should be the one with the earliest expires date, and the highest quantity
    that's still below the user specified (IE: ordered) quantity, that matches a given product.
    """
    product = models.ForeignKey(Product)
    price = CurrencyField(_("Price"), max_digits=14, decimal_places=6)
    quantity = models.DecimalField(_("Discount Quantity"), max_digits=18, 
        decimal_places=6, default='1.0', 
        help_text=_("Use this price only for this quantity or higher"))
    expires = models.DateField(_("Expires"), null=True, blank=True)
    #TODO: add fields here for locale/currency specific pricing

    def __unicode__(self):
        return unicode(self.price)
        
    def adjustments(self):
        """Get a list of price adjustments, in the form of a PriceAdjustmentCalc object.
        """
        
        adjust = PriceAdjustmentCalc(self)
        signals.satchmo_price_query.send(self, adjustment=adjust, 
            slug=self.product.slug, discountable=self.product.is_discountable)
        return adjust

    def _dynamic_price(self):
        """Get the current price as modified by all listeners."""
        
        adjustment = self.adjustments()
        return adjustment.final_price()

    dynamic_price = property(fget=_dynamic_price)

    def save(self, **kwargs):
        prices = Price.objects.filter(product=self.product, quantity=self.quantity)
        ## Jump through some extra hoops to check expires - if there's a better way to handle this field I can't think of it. Expires needs to be able to be set to None in cases where there is no expiration date.
        if self.expires:
            prices = prices.filter(expires=self.expires)
        else:
            prices = prices.filter(expires__isnull=True)
        if self.id:
            prices = prices.exclude(id=self.id)
        if prices.count():
            return #Duplicate Price

        super(Price, self).save(**kwargs)
        ProductPriceLookup.objects.smart_create_for_product(self.product)

    class Meta:
        ordering = ['expires', '-quantity']
        verbose_name = _("Price")
        verbose_name_plural = _("Prices")
        unique_together = (("product", "quantity", "expires"),)

class ProductImage(models.Model):
    """
    A picture of an item.  Can have many pictures associated with an item.
    Thumbnails are automatically created.
    """
    product = models.ForeignKey(Product, null=True, blank=True)
    picture = ImageWithThumbnailField(verbose_name=_('Picture'),
        upload_to="__DYNAMIC__",
        name_field="_filename",
        max_length=200) #Media root is automatically prepended
    caption = models.CharField(_("Optional caption"), max_length=100,
        null=True, blank=True)
    sort = models.IntegerField(_("Sort Order"), default=0)

    def translated_caption(self, language_code=None):
        return lookup_translation(self, 'caption', language_code)

    def _get_filename(self):
        if self.product:
            return '%s-%s' % (self.product.slug, self.id)
        else:
            return 'default'
    _filename = property(_get_filename)

    def __unicode__(self):
        if self.product:
            return u"Image of Product %s" % self.product.slug
        elif self.caption:
            return u"Image with caption \"%s\"" % self.caption
        else:
            return u"%s" % self.picture

    class Meta:
        ordering = ['sort']
        verbose_name = _("Product Image")
        verbose_name_plural = _("Product Images")

class ProductImageTranslation(models.Model):
    """A specific language translation for a `ProductImage`.  This is intended for all descriptions which are not the
    default settings.LANGUAGE.
    """
    productimage = models.ForeignKey(ProductImage, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    caption = models.CharField(_("Translated Caption"), max_length=255, )
    version = models.IntegerField(_('version'), default=1)
    active = models.BooleanField(_('active'), default=True)

    class Meta:
        verbose_name = _('Product Image Translation')
        verbose_name_plural = _('Product Image Translations')
        ordering = ('productimage', 'caption','languagecode')
        unique_together = ('productimage', 'languagecode', 'version')

    def __unicode__(self):
        return u"ProductImageTranslation: [%s] (ver #%i) %s Name: %s" % (self.languagecode, self.version, self.productimage, self.name)

class TaxClass(models.Model):
    """
    Type of tax that can be applied to a product.  Tax
    might vary based on the type of product.  In the US, clothing could be
    taxed at a different rate than food items.
    """
    title = models.CharField(_("Title"), max_length=20,
        help_text=_("Displayed title of this tax."))
    description = models.CharField(_("Description"), max_length=30,
        help_text=_("Description of products that would be taxed."))

    def __unicode__(self):
        return self.title

    class Meta:
        verbose_name = _("Tax Class")
        verbose_name_plural = _("Tax Classes")

UNSET = object()

# --------------- helpers ------------------

def lookup_translation(obj, attr, language_code=None, version=-1):
    """Get a translated attribute by language.

    If specific language isn't found, returns the attribute from the base object.
    """
    if not language_code:
        language_code = get_language()

    if not hasattr(obj, '_translationcache'):
        obj._translationcache = {}

    short_code = language_code
    pos = language_code.find('_')
    if pos > -1:
        short_code = language_code[:pos]

    else:
        pos = language_code.find('-')
        if pos > -1:
            short_code = language_code[:pos]

    trans = None
    has_key = obj._translationcache.has_key(language_code)
    if has_key:
        if obj._translationcache[language_code] == None and short_code != language_code:
            return lookup_translation(obj, attr, short_code)

    if not has_key:
        q = obj.translations.filter(
            languagecode__iexact = language_code)

        if q.count() == 0:
            obj._translationcache[language_code] = None

            if short_code != language_code:
                return lookup_translation(obj, attr, language_code=short_code, version=version)

            else:
                q = obj.translations.filter(
                    languagecode__istartswith = language_code)

        if q.count() > 0:
            trans = None
            if version > -1:
                trans = q.order_by('-version')[0]
            else:
                # try to get the requested version, if it is available,
                # else fallback to the most recent version
                fallback = None
                for t in q.order_by('-version'):
                    if not fallback:
                        fallback = t
                    if t.version == version:
                        trans = t
                        break
                if not trans:
                    trans = fallback

            obj._translationcache[language_code] = trans

    if not trans:
        trans = obj._translationcache[language_code]

    if not trans:
        trans = obj

    val = getattr(trans, attr, UNSET)
    if trans != obj and (val in (None, UNSET)):
        val = getattr(obj, attr)

    return mark_safe(val)

def get_product_quantity_adjustments(product, qty=1, parent=None):
    """Gets a list of adjustments for the price found for a product/qty"""

    qty_discounts = product.price_set.exclude(
        expires__isnull=False,
        expires__lt=datetime.date.today()).filter(quantity__lte=qty)

    adjustments = None

    if qty_discounts.count() > 0:
        # Get the price with the quantity closest to the one specified without going over
        adjustments = qty_discounts.order_by('price','-quantity', 'expires')[0].adjustments()

    elif parent:
        adjustments = get_product_quantity_adjustments(parent, qty=qty)

    if not adjustments:
        adjustments = PriceAdjustmentCalc(None)
        
    return adjustments

def get_product_quantity_price(product, qty=Decimal('1'), delta=Decimal("0.00"), parent=None):
    """
    Returns price as a Decimal else None.
    First checks the product, if none, then checks the parent.
    """

    adjustments = get_product_quantity_adjustments(product, qty=qty, parent=parent)
    
    return adjustments.final_price()+delta

def make_option_unique_id(groupid, value):
    return '%s-%s' % (str(groupid), str(value),)

def round_cents(work):
    cents = Decimal("0.01")
    for lid in work:
        work[lid] = work[lid].quantize(cents)

def sorted_tuple(lst):
    ret = []
    for x in lst:
        if not x in ret:
            ret.append(x)
    ret.sort()
    return tuple(ret)

def split_option_unique_id(uid):
    "reverse of make_option_unique_id"

    parts = uid.split('-')
    return (parts[0], '-'.join(parts[1:]))

# -------------------------------------------
# helper objects - not Django model objects

class PriceAdjustmentCalc(object):
    """Helper class to handle adding up product pricing adjustments"""
    
    def __init__(self, price, product=None):
        self.price = price
        self.base_product = product
        self.adjustments = []
        
    def __add__(self, adjustment):
        self.adjustments.append(adjustment)
        return self
        
    def total_adjustment(self):
        total = Decimal(0)
        for adj in self.adjustments:
            total += adj.amount
        return total
        
    def _product(self):
        """Lazy product dereference"""
        if self.base_product:
            product = self.base_product
        else:
            product = self.price.product
        return product
        
    product = property(fget=_product)
    
    def final_price(self):
        total = Decimal(0)
        if self.price:
            total = self.price.price
            if total is None:
                total = Decimal(0)
        return total - self.total_adjustment()
    
class PriceAdjustment(object):
    """A single product pricing adjustment"""
    
    def __init__(self, key, label=None, amount=None):
        if label is None:
            label = key.capitalize()
        if amount is None:
            amount = Decimal(0)
        self.key = key
        self.label = label
        self.amount = amount

    def __unicode__(self):
        return u"%s: %s=%s" % (_('Price Adjustment'), self.label, moneyfmt(self.amount))
        



from payment.modules.giftcertificate.models import GiftCertificate, GiftCertificateUsage, GiftCertificateProduct
from django.contrib import admin
from django.utils.translation import get_language, ugettext_lazy as _


class GiftCertificateUsage_Inline(admin.StackedInline):
    model = GiftCertificateUsage
    extra = 1

class GiftCertificateOptions(admin.ModelAdmin):
    list_display = ['site', 'code','balance']
    list_display_links = ('code',)
    ordering = ['site', 'date_added']
    inlines = [GiftCertificateUsage_Inline]

admin.site.register(GiftCertificate, GiftCertificateOptions)
admin.site.register(GiftCertificateProduct)




# Copyright (C) 2005 by Mattias Karlsson
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import gobject
import gtk
from kiwi.utils import gsignal, type_register
from kiwi.component import implements

from gazpacho.interfaces import IReferencable

class GSizeGroup(gobject.GObject):

    implements(IReferencable)

    gsignal('name-changed')
    gsignal('gadgets-added', object)
    gsignal('gadgets-removed', object)

    def __init__(self, name, group, gadgets=[]):
        """
        Initialize the sizegroup.

        @param name: the name
        @type name: str
        @param group: sizegroup
        @type group: a gtk.SizeGroup
        """
        gobject.GObject.__init__(self)

        if not isinstance(group, gtk.SizeGroup):
            raise TypeError("group must be a gtk.SizeGroup, not %r" % group)

        self._name = name
        self._gtk_sizegroup = group
        self._gadgets = gadgets or [] # XXX: fixme

    #
    # Properties
    #
    def _set_name(self, name):
        """
        Set the sizegroup name.

        @param name: the name
        @type name: str
        """
        self._name = name
        self.emit('name_changed')

    def _get_name(self):
        """
        Get the sizegroup name.

        @return: the sizegroup name
        @rtype: str
        """
        return self._name

    name = property(_get_name, _set_name)

    def _set_mode(self, mode):
        """
        Set the sizegroup mode.

        @param mode: the sizegroup mode
        @type mode: constants, see gtk.SizeGroup
        """
        self._gtk_sizegroup.set_mode(mode)
        self.emit('name-changed')

    def _get_mode(self):
        """
        Get the SizeGroup mode.

        @return: the sizegroup mode
        @rtype: constants, see gtk.SizeGroup
        """
        return self._gtk_sizegroup.get_mode()

    mode = property(_get_mode, _set_mode)

    #
    # Public methods
    #
    def is_empty(self):
        """
        Check if the sizegroup is empty

        @return: true if the sizegroup doesn't have any gadgets
        @rtype: bool
        """
        return not self._gadgets

    def has_gadget(self, gadget):
        """
        Check if the size group has the gadget.

        @param gadget: the gadget to test
        @type gadget: gazpacho.gadget.Widget

        @return: true if the gadget is in the group
        @rtype: bool
        """
        return gadget in self._gadgets

    def get_gadgets(self):
        """
        Get the gadgets for this group.

        @return: the gadgets
        @rtype: list (of gazpacho.gadget.Widget)
        """
        return self._gadgets[:]

    def add_gadgets(self, gadgets):
        """
        Add gadgets to the SizeGroup.

        @param gadgets: the gadgets to add
        @type gadgets: list (of gazpacho.gadget.Widget)
        """
        if not gadgets:
            return

        self._gadgets += gadgets
        for gadget in gadgets:
            gadget.references.add_referrer(self)
            prop = gadget.get_prop('sizegroup')
            prop.value = self._name
            self._add_widget(gadget.widget)

        self.emit('gadgets-added', gadgets)

    def remove_gadgets(self, gadgets):
        """
        Remove gadgets from the sizegroup.

        @param gadgets: the gadgets to remove
        @type gadgets: list (of gazpacho.gadget.Widget)
        """
        if not gadgets:
            return

        for gadget in gadgets:
            gadget.references.remove_referrer(self)
            self._gadgets.remove(gadget)
            self._remove_widget(gadget.widget)
            prop = gadget.get_prop('sizegroup')
            prop.value = None

        self.emit('gadgets-removed', gadgets)

    # IReferencable
 
    def remove_reference(self, gadget):
        """
        See L{gazpacho.interfaces.IReferencable.remove_reference}
        """
        self._gadgets.remove(gadget)
        prop = gadget.get_prop('sizegroup')
        prop.value = None
        self._remove_widget(gadget.widget)
        self.emit('gadgets-removed', [gadget])
 
    def add_reference(self, gadget):
        """
        See L{gazpacho.interfaces.IReferencable.add_reference}
        """
        self._gadgets.append(gadget)
        prop = gadget.get_prop('sizegroup')
        prop.value = self._name
        self._add_widget(gadget.widget)
        self.emit('gadgets-added', [gadget])

    #
    # Private methods
    #
    def _add_widget(self, widget):
        """
        Add a gtk widget to the gtk SizeGroup.

        @param widget: the gtk widget
        @type widget: gtk.Widget
        """
        self._gtk_sizegroup.add_widget(widget)

    def _remove_widget(self, widget):
        """
        Remove a gtk widget from the gtk SizeGroup.

        @param widget: the gtk widget
        @type widget: gtk.Widget
        """
        self._gtk_sizegroup.remove_widget(widget)

type_register(GSizeGroup)

def safe_to_add_gadgets(sizegroup, gadgets):
    """
    Test if it is safe to add the gadgets to the sizegroup. Gadgets
    who has an ancestor or child in the sizegroup cannot be
    added.

    @param sizegroup: the size group
    @type sizegroup: L{gazpacho.sizegroup.GSizeGroup}
    """
    all_gadgets = sizegroup.get_gadgets() + gadgets

    # Create gadget parent cache
    parents = {}
    for gadget in all_gadgets:
        parent = gadget.get_parent()
        while parent:
            if parent in parents:
                break
            parents[parent] = parent
            parent = parent.get_parent()

    for gadget in all_gadgets:
        if gadget in parents:
            return False

    return True



from django import template
from livesettings import config_value
from l10n.utils import moneyfmt
from tax.templatetags.satchmo_tax import CartitemLineTaxedTotalNode, CartTaxedTotalNode
import logging

log = logging.getLogger('shop.templatetags.satchmo_cart')

register = template.Library()

class CartitemTotalNode(template.Node):
    """Show the total for the cartitem"""

    def __init__(self, cartitem, show_currency, show_tax):
        self.cartitem = template.Variable(cartitem)
        self.raw_cartitem = cartitem
        self.show_currency = template.Variable(show_currency)
        self.raw_currency = show_currency
        self.show_tax = template.Variable(show_tax)
        self.raw_tax = show_tax

    def render(self, context):

        try:
            show_tax = self.show_tax.resolve(context)
        except template.VariableDoesNotExist:
            show_tax = self.raw_tax

        if show_tax:
            tag = CartitemLineTaxedTotalNode(self.raw_cartitem, self.raw_currency)
            return tag.render(context)

        try:
            cartitem = self.cartitem.resolve(context)
        except template.VariableDoesNotExist:
            log.warn('Could not resolve template variable: %s', self.cartitem)
            return ''

        try:
            show_currency = self.show_currency.resolve(context)
        except template.VariableDoesNotExist:
            show_currency = self.raw_currency

        if show_currency:
            return moneyfmt(cartitem.line_total)
        else:
            return cartitem.line_total

def cartitem_custom_details(cartitem):
    is_custom = "CustomProduct" in cartitem.product.get_subtypes()

    return {
        'cartitem' : cartitem,
        'is_custom' : is_custom
    }

register.inclusion_tag("product/cart_detail_customproduct.html", takes_context=False)(cartitem_custom_details)

def cartitem_subscription_details(cartitem):
    log.debug('sub details')
    return {
        'cartitem' : cartitem,
        'is_subscription' : cartitem.product.is_subscription
    }

register.inclusion_tag("product/cart_detail_subscriptionproduct.html", takes_context=False)(cartitem_subscription_details)

def cartitem_total(parser, token):
    """Returns the line total for the cartitem, possibly with tax added.  If currency evaluates true,
    then return the total formatted through moneyfmt.

    Example::

        {% cartitem_total cartitem [show_tax] [currency] %}
    """

    tokens = token.contents.split()
    if len(tokens) < 2:
        raise template.TemplateSyntaxError, "'%s' tag requires a cartitem argument" % tokens[0]

    cartitem = tokens[1]

    if len(tokens) > 2:
        show_tax = tokens[2]
    else:
        show_tax = config_value('TAX', 'DEFAULT_VIEW_TAX')

    if len(tokens) >3:
        show_currency = tokens[3]
    else:
        show_currency = 'True'

    return CartitemTotalNode(cartitem, show_currency, show_tax)

cartitem_total = register.tag(cartitem_total)

class CartTotalNode(template.Node):
    """Show the total for the cart"""

    def __init__(self, cart, show_currency, show_tax, show_discount):
        self.cart = template.Variable(cart)
        self.raw_cart = cart
        self.show_currency = template.Variable(show_currency)
        self.raw_currency = show_currency
        self.show_tax = template.Variable(show_tax)
        self.raw_tax = show_tax
        self.show_discount = template.Variable(show_discount)
        self.raw_show_discount = show_discount

    def render(self, context):

        try:
            show_tax = self.show_tax.resolve(context)
        except template.VariableDoesNotExist:
            show_tax = self.raw_tax

        if show_tax:
            tag = CartTaxedTotalNode(self.raw_cart, self.raw_currency)
            return tag.render(context)

        try:
            cart = self.cart.resolve(context)
        except template.VariableDoesNotExist:
            log.warn('Could not resolve template variable: %s', self.cart)
            return ''

        try:
            show_currency = self.show_currency.resolve(context)
        except template.VariableDoesNotExist:
            show_currency = self.raw_currency

        try:
            show_discount = self.show_discount.resolve(context)
        except template.VariableDoesNotExist:
            show_discount = self.raw_show_discount

        if show_discount:
            total = cart.total_undiscounted
        else:
            total = cart.total

        if show_currency:
            return moneyfmt(cart.total)
        else:
            return cart.total

def cart_total(parser, token):
    """Returns the total for the cart, possibly with tax added.  If currency evaluates true,
    then return the total formatted through moneyfmt.

    Example::

        {% cart_total cart [show_tax] [currency] [discounted] %}
    """

    tokens = token.contents.split()
    if len(tokens) < 2:
        raise template.TemplateSyntaxError, "'%s' tag requires a cart argument" % tokens[0]

    cart = tokens[1]

    if len(tokens) > 2:
        show_tax = tokens[2]
    else:
        show_tax = config_value('TAX', 'DEFAULT_VIEW_TAX')

    if len(tokens) > 3:
        show_currency = tokens[3]
    else:
        show_currency = True

    if len(tokens) > 4:
        show_discount = tokens[4]
    else:
        show_discount = False

    return CartTotalNode(cart, show_currency, show_tax)

cart_total = register.tag(cart_total)



# Copyright (C) 2004,2005 by SICEm S.L.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import xml.dom.minidom

import gtk
import gobject

from gazpacho import util
from gazpacho.i18n import _
from gazpacho.loader import tags

class Converter:
    def __init__(self, xml_node, uimanager, ui_definitions, signals):
        self.root_node = xml_node
        self.uimanager = uimanager
        self.ui_definitions = ui_definitions
        self.signals = signals
        self.widget_id = xml_node.getAttribute(tags.XML_TAG_ID)

    def build_widget(self):
        """This should be implemented by subclasses"""

class UIManagerBasedWidgetsConverter(Converter):
    """Superclass for converters of widgets based on UIManager like
    Toolbars and MenuBars
    """
    root_node_name = '' # subclasses should override this attribute

    def __init__(self, xml_node, uimanager, ui_definitions, signals):
        Converter.__init__(self, xml_node, uimanager, ui_definitions, signals)
        self.action_group = gtk.ActionGroup('Deprecated Actions for %s' % \
                                            self.widget_id)
        self.uimanager.insert_action_group(self.action_group, 0)

    def build_widget(self):
        new_doc = xml.dom.getDOMImplementation().createDocument(None, None,
                                                                None)
        root_node = new_doc.createElement('ui')
        new_doc.appendChild(root_node)

        widget_node = new_doc.createElement(self.root_node_name)
        root_node.appendChild(widget_node)

        widget_node.setAttribute('action', self.widget_id)
        widget_node.setAttribute('name', self.widget_id)

        for child_node in util.xml_child_widgets(self.root_node):
            self.build_child_widget(child_node, widget_node, new_doc)

        xml_string = new_doc.documentElement.toxml()
        merge_id = self.uimanager.add_ui_from_string(xml_string)
        obj = self.uimanager.get_widget('ui/%s' % self.widget_id)
        obj.set_data('merge_id', merge_id)

        if 'initial-state' in self.ui_definitions:
            other_xml_string = self.ui_definitions['initial-state']
            other_xml = xml.dom.minidom.parseString(other_xml_string)
            for node in [n for n in other_xml.childNodes if n.typeNode == n.ELEMENT_NODE]:
                new_doc.appendChild(node.cloneNode())
            xml_string = new_doc.documentElement.toxml()

        self.ui_definitions['initial-state'] = xml_string
        return obj

    def build_child_widget(self, xml_node, parent_node, new_doc):
        """This should be implemented by subclasses"""

    def create_action_from_node(self, xml_node, action_name):
        props = {'label' : action_name, 'tooltip': '', 'stock-id': None}
        for child_node in util.xml_filter_nodes(xml_node.childNodes,
                                                xml_node.ELEMENT_NODE):
            if child_node.tagName == tags.XML_TAG_PROPERTY:
                name, value = util.xml_parse_raw_property(child_node)
                if name == 'label':
                    props[name] = value
            elif child_node.tagName == tags.XML_TAG_SIGNAL:
                signal_name = child_node.getAttribute(tags.XML_TAG_NAME)
                if signal_name in ('clicked', 'activated'):
                    callback = child_node.getAttribute(tags.XML_TAG_HANDLER)
                    props['callback'] = callback

        action = gtk.Action(action_name, props['label'], props['tooltip'],
                            props['stock-id'])
        if 'callback' in list(props.keys()):
            if props['callback']:
                action.set_data(tags.SIGNAL_HANDLER, props['callback'])
                self.signals.append([action, 'activate', callback, None])
        self.action_group.add_action(action)

class ToolbarConverter(UIManagerBasedWidgetsConverter):
    root_node_name = 'toolbar'

    def build_child_widget(self, xml_node, parent_node, new_doc):
        if xml_node.tagName != tags.XML_TAG_WIDGET:
            return

        item_id = xml_node.getAttribute(tags.XML_TAG_ID)
        class_name = xml_node.getAttribute(tags.XML_TAG_CLASS)
        if class_name == 'GtkSeparatorToolItem':
            tool_node = new_doc.createElement('separator')
        else:
            tool_node = new_doc.createElement('toolitem')
            tool_node.setAttribute('action', item_id)
            tool_node.setAttribute('name', item_id)
            self.create_action_from_node(xml_node, item_id)

        parent_node.appendChild(tool_node)

class MenuBarConverter(UIManagerBasedWidgetsConverter):

    root_node_name = 'menubar'

    def build_child_widget(self, xml_node, parent_node, new_doc):
        if xml_node.tagName != tags.XML_TAG_WIDGET:
            return
        item_id = xml_node.getAttribute(tags.XML_TAG_ID)
        menu_node = new_doc.createElement('menu')
        menu_node.setAttribute('action', item_id)
        menu_node.setAttribute('name', item_id)
        parent_node.appendChild(menu_node)

        self.create_action_from_node(xml_node, item_id)

        for child_node in util.xml_child_widgets(xml_node):
            # this is the GtkMenu node. we skip top level GtkMenu nodes
            for subchild_node in util.xml_child_widgets(child_node):
                self.build_menu_item(subchild_node, menu_node, new_doc)

    def build_menu_item(self, xml_node, parent_node, new_doc):
        if xml_node.tagName != tags.XML_TAG_WIDGET:
            return
        item_id = xml_node.getAttribute(tags.XML_TAG_ID)
        # if the menuitem has children it is actually a menu
        children = util.xml_child_widgets(xml_node)
        if children:
            menu_node = new_doc.createElement('menu')
            menu_node.setAttribute('action', item_id)
            menu_node.setAttribute('name', item_id)
            self.create_action_from_node(xml_node, item_id)
            for child_node in util.xml_child_widgets(xml_node):
                self.build_menu_item(child_node, menu_node, new_doc)

        else:
            class_name = xml_node.getAttribute(tags.XML_TAG_CLASS)
            if class_name == 'GtkSeparatorMenuItem':
                menu_node = new_doc.createElement('separator')
            else:
                menu_node = new_doc.createElement('menuitem')
                menu_node.setAttribute('action', item_id)
                menu_node.setAttribute('name', item_id)
                self.create_action_from_node(xml_node, item_id)

        parent_node.appendChild(menu_node)

builders = {
    'GtkMenuBar': MenuBarConverter,
    'GtkToolbar': ToolbarConverter,
    }

def build_deprecated_widget(xml_node, uimanager, ui_definitions, signals):
    global builders
    klass = xml_node.getAttribute(tags.XML_TAG_CLASS)
    obj = builders[klass](xml_node, uimanager, ui_definitions, signals)
    return obj.build_widget()


class DeprecatedWidgetsDialog(gtk.Dialog):
    def __init__(self, window, widgets):
        gtk.Dialog.__init__(self,
                            title=_('Deprecated widgets found'),
                            parent=window,
                            flags=gtk.DIALOG_MODAL | \
                            gtk.DIALOG_DESTROY_WITH_PARENT | \
                            gtk.DIALOG_NO_SEPARATOR,
                            buttons=(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))

        self.set_border_width(12)
        self.vbox.set_spacing(6)

        msg = _('Gazpacho found the following deprecated widgets:')
        label = gtk.Label('<b><big>%s</big></b>' % msg)
        label.set_use_markup(True)
        self.vbox.pack_start(label, False)

        msg = '\n'.join(["- %s: %s" % (w.name, c) for w, c in widgets])
        label = gtk.Label(msg)
        self.vbox.pack_start(label, False)

        msg = _("They were converted to the equivalent newest widgets "
                "but please note that if\nyou save the file in Gazpacho "
                "it will not be compatible with Glade-2 anymore")
        label = gtk.Label('<small>%s</small>' % msg)
        label.set_use_markup(True)
        self.vbox.pack_start(label, False)

        self.vbox.show_all()

gobject.type_register(DeprecatedWidgetsDialog)



from django.contrib.sites.models import Site
from django.core import urlresolvers
from satchmo_utils import url_join
from django.contrib.sites.models import Site
import logging
log = logging.getLogger('satchmo_utils')

def lookup_template(settings, template):
    """Return a template name, which may have been overridden in the settings."""

    if settings.has_key('TEMPLATE_OVERRIDES'):
        val = settings['TEMPLATE_OVERRIDES']
        template = val.get(template, template)

    return template

def lookup_url(settings, name, include_server=False, ssl=False):
    """Look up a named URL for the payment module.

    Tries a specific-to-general lookup fallback, returning
    the first positive hit.

    First look for a dictionary named "URL_OVERRIDES" on the settings object.
    Next try prepending the module name to the name
    Last just look up the name
    """
    url = None

    if settings.has_key('URL_OVERRIDES'):
        val = settings['URL_OVERRIDES']
        url = val.get(name, None)

    if not url:
        try:
            possible = settings.KEY.value + "_" + name
            url = urlresolvers.reverse(possible)
        except urlresolvers.NoReverseMatch:
            log.debug('No url found for %s', possible)

    if not url:
        try:
            url = urlresolvers.reverse(name)
        except urlresolvers.NoReverseMatch:
            log.error('Could not find any url for %s', name)
            raise urlresolvers.NoReverseMatch('No reverse for %s or %s' % (possible, name))
            
    if include_server:
        if ssl:
            method = "https://"
        else:
            method = "http://"
        site = Site.objects.get_current()
        url = url_join(method, site.domain, url)

    return url



import re

from django.utils.translation import ugettext_lazy as _

US_POSTCODE_RE = re.compile(r'^\d{5}(?:-\d{4})?$')
def validate(postcode):
    if US_POSTCODE_RE.search(postcode):
        return postcode
    else:
        raise ValueError(_('Invalid ZIP code'))



from decimal import Decimal
from django.utils.translation import ugettext_lazy as _
from livesettings import *

SHIP_MODULES = config_get('SHIPPING', 'MODULES')
SHIP_MODULES.add_choice(('shipping.modules.usps', 'USPS'))

SHIPPING_GROUP = ConfigurationGroup('shipping.modules.usps',
    _('U.S.P.S. Shipping Settings'),
    requires = SHIP_MODULES,
    ordering = 101)

config_register_list(

StringValue(SHIPPING_GROUP,
    'USER_ID',
    description=_("USPS  Username"),
    help_text=_("User ID provided by USPS site."),
    default=u""),

StringValue(SHIPPING_GROUP,
    'USER_PASSWORD',
    description=_("USPS Password"),
    help_text=_("User password provided by USPS site."),
    default=u""),

DecimalValue(SHIPPING_GROUP,
    'HANDLING_FEE',
    description=_("Handling Fee"),
    help_text=_("The cost of packaging and taking order to post office"),
    default=Decimal('0.00')),

MultipleStringValue(SHIPPING_GROUP,
    'USPS_SHIPPING_CHOICES',
    description=_("USPS Shipping Choices Available to customers."),
    choices = (
        (('0', 'First Class')),
        (('1', 'Priority Mail')),
        (('16', 'Priority Mail Flat-Rate Envelope')),
        (('17', 'Priority Mail Flat-Rate Box')),
        (('22', 'Priority Mail Large Flat-Rate Box')),
        (('3', 'Express Mail')),
        (('13', 'Express Mail Flat-Rate Envelope')),
        (('4', 'Parcel Post')),
        (('5', 'Bound Printed Matter')),
        (('6', 'Media Mail')),
        (('7', 'Library Mail')),
        
        # INTERNATIONAL CODES
        
        (('14', 'Int`l: First Class Mail International Large Envelope')),
        (('15', 'Int`l: First Class Mail International Package')),
        (('1', 'Int`l: Express Mail International (EMS)')),
        (('4', 'Int`l: Global Express Guaranteed')),
        (('6', 'Int`l: Global Express Guaranteed Non-Document Rectangular')),
        (('7', 'Int`l: Global Express Guaranteed Non-Document Non-Rectangular')),
        (('10', 'Int`l: Express Mail International (EMS) Flat-Rate Envelope')),
        (('2', 'Int`l: Priority Mail International')),
        (('8', 'Int`l: Priority Mail International Flat-Rate Envelope')),
        (('9', 'Int`l: Priority Mail International Flat-Rate Box')),
        (('11', 'Int`l: Priority Mail International Large Flat-Rate Box')),
        (('12', 'Int`l: USPS GXG Envelopes')),
    ),
    default = ('3',)),

StringValue(SHIPPING_GROUP,
    'SHIPPING_CONTAINER',
    description=_("Type of container used to ship product."),
    choices = (
        (('00', 'Unknown')),
        (('01', 'Variable')),
        (('02', 'Flat rate box')),
        (('03', 'Flat rate envelope')),
        (('04', 'Rectangular')),
        (('05', 'Non-rectangular')),
    ),
    default = u"01"),

BooleanValue(SHIPPING_GROUP,
    'LIVE',
    description=_("Access production USPS server"),
    help_text=_("Use this when your store is in production."),
    default=False),

StringValue(SHIPPING_GROUP,
    'CONNECTION',
    description=_("Submit to URL"),
    help_text=_("Address to submit live transactions."),
    default="http://production.shippingapis.com/ShippingAPI.dll"),

StringValue(SHIPPING_GROUP,
    'CONNECTION_TEST',
    description=_("Submit to Test URL"),
    help_text=_("Address to submit test transactions."),
    default="http://testing.shippingapis.com/ShippingAPITest.dll"),

BooleanValue(SHIPPING_GROUP,
    'VERBOSE_LOG',
    description=_("Verbose logs"),
    help_text=_("Send the entire request and response to the log - for debugging help when setting up USPS."),
    default=False),

#ModuleValue(SHIPPING_GROUP,
#    'MODULE',
#    description=_('Implementation module'),
#    hidden=True,
#    default = 'shipping.modules.ups'),

)



from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models
from django.utils.translation import ugettext_lazy as _
from l10n.mixins import TranslatedObjectMixin
import product
from product.models import Product
from satchmo_utils.thumbnail.field import ImageWithThumbnailField
from signals_ahoy.signals import collect_urls
import logging

log = logging.getLogger('brand.models')

class BrandManager(models.Manager):
    
    def active(self, site=None):
        if not site:
            site = Site.objects.get_current()
        return self.filter(site=site, active=True)
    
    def by_slug(self, slug):
        site = Site.objects.get_current()
        return self.get(slug=slug, site=site)

class Brand(models.Model, TranslatedObjectMixin):
    """A product brand"""
    site = models.ForeignKey(Site)
    slug = models.SlugField(_("Slug"), unique=True,
    help_text=_("Used for URLs"))
    products = models.ManyToManyField(Product, blank=True, verbose_name=_("Products"), through='BrandProduct')
    ordering = models.IntegerField(_("Ordering"))
    active = models.BooleanField(default=True)
    
    objects = BrandManager()
    
    def _active_categories(self):
        return [cat for cat in self.categories.all() if cat.has_content()]
    
    active_categories = property(fget=_active_categories)
    
    def _translation(self):
        return self._find_translation()
    translation = property(fget=_translation)

    def _get_absolute_url(self):
        return ('satchmo_brand_view', None, {'brandname' : self.slug})
        
    get_absolute_url = models.permalink(_get_absolute_url)
        
    def active_products(self):
        return self.products.filter(site=self.site, active=True)        

    def has_categories(self):
        return self.active_categories().count() > 0

    def has_content(self):
        return self.has_products() or self.has_categories()

    def has_products(self):
        return self.active_products().count() > 0
            
    def __unicode__(self):
        return u"%s" % self.slug
            
    class Meta:
        ordering=('ordering', 'slug')
        verbose_name = _('Brand')
        verbose_name_plural = _('Brands')

class BrandProduct(models.Model):
    brand = models.ForeignKey(Brand)
    product = models.ForeignKey(Product)
    
    class Meta:
        verbose_name=_("Brand Product")
        verbose_name_plural=_("Brand Products")

class BrandTranslation(models.Model):
    brand = models.ForeignKey(Brand, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    name = models.CharField(_('title'), max_length=100, blank=False)
    short_description = models.CharField(_('Short Description'), blank=True, max_length=200)
    description = models.TextField(_('Full Description'), blank=True)
    picture = ImageWithThumbnailField(verbose_name=_('Picture'),
        upload_to="__DYNAMIC__",
        name_field="_filename",
        null=True, blank=True,
        max_length=200) #Media root is automatically prepended
    
    def _get_filename(self):
        if self.brand:
            return '%s-%s' % (self.brand.slug, self.id)
        else:
            return 'default'
    _filename = property(_get_filename)

    class Meta:
        ordering=('languagecode', )      
        verbose_name = _('Brand Translation')
        verbose_name_plural = _('Brand Translations')  

class BrandCategoryManager(models.Manager):
    def by_slug(self, brandname, slug):
        brand = Brand.objects.by_slug(brandname)
        return brand.categories.get(slug=slug)

class BrandCategory(models.Model, TranslatedObjectMixin):
    """A category within a brand"""
    slug = models.SlugField(_("Slug"),
        help_text=_("Used for URLs"))
    brand = models.ForeignKey(Brand, related_name="categories")
    products = models.ManyToManyField(Product, blank=True, verbose_name=_("Products"), through='BrandCategoryProduct')
    ordering = models.IntegerField(_("Ordering"))
    active = models.BooleanField(default=True)

    objects = BrandCategoryManager()

    def _translation(self):
        return self._find_translation()
    translation = property(fget=_translation)

    def _get_absolute_url(self):
        return ('satchmo_brand_category_view', None, {'brandname' : self.brand.slug, 'catname' : self.slug})
    
    get_absolute_url = models.permalink(_get_absolute_url)
        
    def active_products(self):
        return self.products.filter(site=self.brand.site).filter(active=True)                
        
    def has_categories(self):
        return False    
    
    def has_content(self):
        return self.active_products()

    def has_products(self):
        return self.active_products().count > 0

    def __unicode__(self):
        return u"%s: %s" % (self.brand.slug, self.slug)

    class Meta:
        ordering=('ordering', 'slug')
        verbose_name = _('Brand Category')
        verbose_name_plural = _('Categories')

class BrandCategoryProduct(models.Model):
    brandcategory = models.ForeignKey(BrandCategory)
    product = models.ForeignKey(Product)
    
    class Meta:
        verbose_name = _('Brand Category Product')
        verbose_name_plural = _('Brand Category Products')

class BrandCategoryTranslation(models.Model):

    brandcategory = models.ForeignKey(BrandCategory, related_name="translations")
    languagecode = models.CharField(_('language'), max_length=10, choices=settings.LANGUAGES)
    name = models.CharField(_('title'), max_length=100, blank=False)
    short_description = models.CharField(_('Short Description'), blank=True, max_length=200)
    description = models.TextField(_('Description'), blank=True)
    picture = ImageWithThumbnailField(verbose_name=_('Picture'),
        upload_to="__DYNAMIC__",
        name_field="_filename",
        null=True, blank=True,
        max_length=200) #Media root is automatically prepended
    
    def _get_filename(self):
        if self.brandcategory:
            return '%s-%s' % (self.brandcategory.brand.slug, self.id)
        else:
            return 'default'
    _filename = property(_get_filename)
    
    class Meta:
        ordering=('languagecode', )
        verbose_name_plural = _('Brand Category Translations')

#import config        
from urls import add_brand_urls
collect_urls.connect(add_brand_urls, sender=product)



"""Type safe JSON serialization for use when round-trip serialization-deserialization is not neccessary."""

# This is code from http://dpaste.com/hold/4601
# Originally by Wolfram Kriesing (http://wolfram.kriesing.de/blog/)
# Used by written permission to license under the same modified 
# BSD license as the rest of Satchmo.

import types
from django.db import models
try:
    from django.utils import simplejson as json
except ImportError:
    import simplejson as json
from django.core.serializers.json import DateTimeAwareJSONEncoder
from decimal import *

def json_encode(data):
    """
    The main issues with django's default json serializer is that properties that
    had been added to a object dynamically are being ignored (and it also has 
    problems with some models).
    """

    def _any(data):
        ret = None
        if type(data) is types.ListType:
            ret = _list(data)
        elif type(data) is types.DictType:
            ret = _dict(data)
        elif isinstance(data, Decimal):
            # json.dumps() cant handle Decimal
            ret = "%d" % data
        elif isinstance(data, models.query.QuerySet):
            # Actually its the same as a list ...
            ret = _list(data)
        elif isinstance(data, models.Model):
            ret = _model(data)
        else:
            ret = data
        return ret
    
    def _model(data):
        ret = {}
        # If we only have a model, we only want to encode the fields.
        for f in data._meta.fields:
            ret[f.attname] = _any(getattr(data, f.attname))
        # And additionally encode arbitrary properties that had been added.
        fields = dir(data.__class__) + ret.keys()
        add_ons = [k for k in dir(data) if k not in fields]
        for k in add_ons:
            ret[k] = _any(getattr(data, k))
        return ret
    
    def _list(data):
        ret = []
        for v in data:
            ret.append(_any(v))
        return ret
    
    def _dict(data):
        ret = {}
        for k,v in data.items():
            ret[k] = _any(v)
        return ret
    
    ret = _any(data)
    
    return json.dumps(ret, cls=DateTimeAwareJSONEncoder)




'''
An encryption plugin for Elixir utilizing the excellent PyCrypto library, which
can be downloaded here: http://www.amk.ca/python/code/crypto

Values for columns that are specified to be encrypted will be transparently
encrypted and safely encoded for storage in a unicode column using the powerful
and secure Blowfish Cipher using a specified "secret" which can be passed into
the plugin at class declaration time.

Example usage:

.. sourcecode:: python

    from elixir import *
    from elixir.ext.encrypted import acts_as_encrypted

    class Person(Entity):
        name = Field(Unicode)
        password = Field(Unicode)
        ssn = Field(Unicode)
        acts_as_encrypted(for_fields=['password', 'ssn'],
                          with_secret='secret')

The above Person entity will automatically encrypt and decrypt the password and
ssn columns on save, update, and load.  Different secrets can be specified on
an entity by entity basis, for added security.

**Important note**: instance attributes are encrypted in-place. This means that
if one of the encrypted attributes of an instance is accessed after the
instance has been flushed to the database (and thus encrypted), the value for
that attribute will be crypted in the in-memory object in addition to the
database row.
'''

from Crypto.Cipher import Blowfish
from elixir.statements import Statement
from sqlalchemy.orm import MapperExtension, EXT_CONTINUE, EXT_STOP

try:
    from sqlalchemy.orm import EXT_PASS
    SA05orlater = False
except ImportError:
    SA05orlater = True

__all__ = ['acts_as_encrypted']
__doc_all__ = []


#
# encryption and decryption functions
#

def encrypt_value(value, secret):
    return Blowfish.new(secret, Blowfish.MODE_CFB) \
                   .encrypt(value).encode('string_escape')

def decrypt_value(value, secret):
    return Blowfish.new(secret, Blowfish.MODE_CFB) \
                   .decrypt(value.decode('string_escape'))


#
# acts_as_encrypted statement
#

class ActsAsEncrypted(object):

    def __init__(self, entity, for_fields=[], with_secret='abcdef'):

        def perform_encryption(instance, encrypt=True):
            encrypted = getattr(instance, '_elixir_encrypted', None)
            if encrypted is encrypt:
                # skipping encryption or decryption, as it is already done
                return
            else:
                # marking instance as already encrypted/decrypted
                instance._elixir_encrypted = encrypt

            if encrypt:
                func = encrypt_value
            else:
                func = decrypt_value

            for column_name in for_fields:
                current_value = getattr(instance, column_name)
                if current_value:
                    setattr(instance, column_name,
                            func(current_value, with_secret))

        def perform_decryption(instance):
            perform_encryption(instance, encrypt=False)

        class EncryptedMapperExtension(MapperExtension):

            def before_insert(self, mapper, connection, instance):
                perform_encryption(instance)
                return EXT_CONTINUE

            def before_update(self, mapper, connection, instance):
                perform_encryption(instance)
                return EXT_CONTINUE

            if SA05orlater:
                def reconstruct_instance(self, mapper, instance):
                    perform_decryption(instance)
                    # no special return value is required for
                    # reconstruct_instance, but you never know...
                    return EXT_CONTINUE
            else:
                def populate_instance(self, mapper, selectcontext, row,
                                      instance, *args, **kwargs):
                    mapper.populate_instance(selectcontext, instance, row,
                                             *args, **kwargs)
                    perform_decryption(instance)
                    # EXT_STOP because we already did populate the instance and
                    # the normal processing should not happen
                    return EXT_STOP

        # make sure that the entity's mapper has our mapper extension
        entity._descriptor.add_mapper_extension(EncryptedMapperExtension())


acts_as_encrypted = Statement(ActsAsEncrypted)




from django.db import models
from django.utils.translation import ugettext_lazy as _
from product.models import Product, get_product_quantity_price, get_product_quantity_adjustments
from decimal import Decimal
import datetime
from satchmo_utils import add_month
from satchmo_utils.fields import CurrencyField

SATCHMO_PRODUCT=True

def get_product_types():
    return ('SubscriptionProduct',)

class SubscriptionProduct(models.Model):
    """
    This type of Product is for recurring billing (memberships, subscriptions, payment terms)
    """
    product = models.OneToOneField(Product, verbose_name=_("Product"), primary_key=True)
    recurring = models.BooleanField(_("Recurring Billing"), help_text=_("Customer will be charged the regular product price on a periodic basis."), default=False)
    recurring_times = models.IntegerField(_("Recurring Times"), help_text=_("Number of payments which will occur at the regular rate.  (optional)"), null=True, blank=True)
    expire_length = models.IntegerField(_("Duration"), help_text=_("Length of each billing cycle"), null=True, blank=True)
    SUBSCRIPTION_UNITS = (
        ('DAY', _('Days')),
        ('MONTH', _('Months'))
    )
    expire_unit = models.CharField(_("Expire Unit"), max_length=5, choices=SUBSCRIPTION_UNITS, default="DAY", null=False)
    SHIPPING_CHOICES = (
        ('0', _('No Shipping Charges')),
        ('1', _('Pay Shipping Once')),
        ('2', _('Pay Shipping Each Billing Cycle')),
    )
    is_shippable = models.IntegerField(_("Shippable?"), help_text=_("Is this product shippable?"), max_length=1, choices=SHIPPING_CHOICES)

    is_subscription = True

    def _get_subtype(self):
        return 'SubscriptionProduct'

    def __unicode__(self):
        return self.product.slug

    def _get_fullPrice(self):
        """
        returns price as a Decimal
        """
        return self.get_qty_price(1)

    unit_price = property(_get_fullPrice)

    def get_qty_price(self, qty, show_trial=True, include_discount=True):
        """
        If QTY_DISCOUNT prices are specified, then return the appropriate discount price for
        the specified qty.  Otherwise, return the unit_price
        returns price as a Decimal

        Note: If a subscription has a trial, then we'll return the first trial price, otherwise the checkout won't
        balance and it will look like there are items to be paid on the order.
        """
        if show_trial:
            trial = self.get_trial_terms(0)
        else:
            trial = None

        if trial:
            price = trial.price * qty
        else:
            if include_discount:
                price = get_product_quantity_price(self.product, qty)
            else:
                adjustment = get_product_quantity_adjustments(self, qty)
                if adjustment.price is not None:
                    price = adjustment.price.price
                else:
                    price = None

            if not price and qty == Decimal('1'):      # Prevent a recursive loop.
                price = Decimal("0.00")
            elif not price:
                price = self.product._get_fullPrice()
        return price

    def recurring_price(self):
        """
        Get the non-trial price.
        """
        return self.get_qty_price(Decimal('1'), show_trial=False)

    # use order_success() and DownloadableProduct.create_key() to add user to group and perform other tasks
    def get_trial_terms(self, trial=None):
        """Get the trial terms for this subscription"""
        if trial is None:
            return self.trial_set.all().order_by('id')
        else:
            try:
                return self.trial_set.all().order_by('id')[trial]
            except IndexError:
                return None

    def calc_expire_date(self, date=None):
        if date is None:
            date = datetime.datetime.now()
        if self.expire_unit == "DAY":
            expiredate = date + datetime.timedelta(days=self.expire_length)
        else:
            expiredate = add_month(date, n=self.expire_length)

        return expiredate

    def save(self, **kwargs):
        if hasattr(self.product,'_sub_types'):
            del self.product._sub_types
        super(SubscriptionProduct, self).save(**kwargs)

    class Meta:
        verbose_name = _("Subscription Product")
        verbose_name_plural = _("Subscription Products")

class Trial(models.Model):
    """
    Trial billing terms for subscription products.
    Separating it out lets us have as many trial periods as we want.
    Note that some third party payment processors support only a limited number of trial
    billing periods.  For example, PayPal limits us to 2 trial periods, so if you are using
    PayPal for a billing option, you need to create no more than 2 trial periods for your
    product.  However, gateway based processors like Authorize.net can support as many
    billing periods as you wish.
    """
    subscription = models.ForeignKey(SubscriptionProduct)
    price = CurrencyField(_("Price"), help_text=_("Set to 0 for a free trial.  Leave empty if product does not have a trial."), max_digits=10, decimal_places=2, null=True, )
    expire_length = models.IntegerField(_("Trial Duration"), help_text=_("Length of trial billing cycle.  Leave empty if product does not have a trial."), null=True, blank=True)

    def __unicode__(self):
        return unicode(self.price)

    def _occurrences(self):
        if self.expire_length:
            return int(self.expire_length/self.subscription.expire_length)
        else:
            return 0
    occurrences = property(fget=_occurrences)

    def calc_expire_date(self, date=None):
        if date is None:
            date = datetime.datetime.now()
        if self.subscription.expire_unit == "DAY":
            expiredate = date + datetime.timedelta(days=self.expire_length)
        else:
            expiredate = add_month(date, n=self.expire_length)

        return expiredate

    class Meta:
        ordering = ['-id']
        verbose_name = _("Trial Terms")
        verbose_name_plural = _("Trial Terms")




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2007-2019 Etienne Robillard
# All rights reserved. 
# <LICENSE=APACHEV2> 
import os
import posixpath
from time import gmtime, strftime
from docutils.core import publish_parts

__all__ = ('HTMLPublisher', )

#RestructuredTextPublisher
class HTMLPublisher(object):
    """ 
    HTML Publisher class supporting reStructuredText source input.
    """

    def __init__(self, source):
        if os.access(source, os.R_OK):
            self.instream = source
        elif isinstance(source, (basestring, str)): # handle UTF-8 and str types
            self.instream = source
        else:
            raise ValueError('Unrecognized source document: %r' % source)
  
        #self.metadata = self.getdocinfo(source)

        self.parts = None
 
    def getdocinfo(self, doc):
        """Return gmtime/mtime (unix timestamp) for ``doc``"""
        gmt = gmtime(posixpath.getmtime(doc))
        mtime = strftime("%Y/%m/%d", gmt)
        
        return (gmt, mtime)

    def render_to_html(self, keyname='html_body', writer_name='html'):
        """Renders the HTML element from the source document"""
        try:
            parts = publish_parts(source=self.instream, writer_name=writer_name)
        except KeyError:
            raise
        return parts[keyname]
        
    #ctx = dict(content=parts['html_body'], mtime=mtime)




from django import http
from django.core.paginator import Paginator, InvalidPage
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from livesettings import config_value
from product.models import Product
from product.queries import bestsellers
import logging
        
log = logging.getLogger('product.views.filters')
    
def display_bestsellers(request, count=0, template='product/best_sellers.html'):
    """Display a list of the products which have sold the most"""
    if count == 0:
        count = config_value('PRODUCT','NUM_PAGINATED')
    
    ctx = RequestContext(request, {
        'products' : bestsellers(count),
    })
    return render_to_response(template, context_instance=ctx)
        
def display_recent(request, page=0, count=0, template='product/recently_added.html'):
    """Display a list of recently added products."""
    if count == 0:
        count = config_value('PRODUCT','NUM_PAGINATED')

    if page == 0:
        if request.method == 'GET':
            page = request.GET.get('page', 1)
        else:
            page = 1
     
    query = Product.objects.recent_by_site()
    paginator = Paginator(query, count)
    try:
        currentpage = paginator.page(page)
    except InvalidPage:
        currentpage = None
    
    ctx = RequestContext(request, {
        'page' : currentpage,
        'paginator' : paginator,
    })
    return render_to_response(template, context_instance=ctx)



try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from decimal import Decimal
from django import forms
from django.conf import settings
from django.contrib.sites.models import Site
from django.core import serializers, urlresolvers
from django.core.management.base import CommandError
from django.core.management.color import no_style
from django.db import transaction
from django.http import HttpResponse
from django.utils.translation import ugettext as _
from livesettings import config_value
from product.models import Product, Price, Option
from satchmo_utils.unique_id import slugify
import config
import logging
import os
import time
import zipfile

log = logging.getLogger('product.forms')

def export_choices():
    fmts = serializers.get_serializer_formats()
    return zip(fmts,fmts)

class InventoryForm(forms.Form):

    def __init__(self, *args, **kwargs):
        products = kwargs.pop('products', None)

        super(InventoryForm, self).__init__(*args, **kwargs)

        if not products:
            products = Product.objects.by_site().order_by('slug')

        for product in products:
            subtypes = product.get_subtypes()
            qtyclasses = ('text', 'qty') + subtypes
            qtyclasses = " ".join(qtyclasses)

            kw = { 
            'label' : product.slug,
            'help_text' : product.name,
            'initial' : product.items_in_stock,
            'widget' : forms.TextInput(attrs={'class': qtyclasses}) }

            qty = forms.DecimalField(**kw)
            self.fields['qty__%s' % product.slug] = qty
            qty.slug = product.slug
            qty.product_id = product.id
            qty.subtypes = " ".join(subtypes)

            if 'CustomProduct' in subtypes:
                initial_price = product.customproduct.full_price
            else:
                initial_price = product.unit_price

            kw['initial'] = initial_price
            kw['required'] = False
            kw['widget'] = forms.TextInput(attrs={'class': "text price"})
            price = forms.DecimalField(**kw)
            price.slug = product.slug
            self.fields['price__%s' % product.slug] = price

            kw['initial'] = product.active
            kw['widget'] = forms.CheckboxInput(attrs={'class': "checkbox active"})
            active = forms.BooleanField(**kw)
            active.slug = product.slug
            self.fields['active__%s' % product.slug] = active

            kw['initial'] = product.featured
            kw['widget'] = forms.CheckboxInput(attrs={'class': "checkbox featured"})
            featured = forms.BooleanField(**kw)
            featured.slug = product.slug
            self.fields['featured__%s' % product.slug] = featured

    def save(self, request):
        self.full_clean()
        for name, value in self.cleaned_data.items():
            opt, key = name.split('__')

            prod = Product.objects.get(slug__exact=key)
            subtypes = prod.get_subtypes()

            if opt=='qty':
                if value != prod.items_in_stock:
                    request.user.message_set.create(message='Updated %s stock to %s' % (key, value))
                    log.debug('Saving new qty=%d for %s' % (value, key))
                    prod.items_in_stock = value
                    prod.save()

            elif opt=='price':
                if 'CustomProduct' in subtypes:
                    full_price = prod.customproduct.full_price
                else:
                    full_price = prod.unit_price

                if value != full_price:
                    request.user.message_set.create(message='Updated %s unit price to %s' % (key, value))
                    log.debug('Saving new price %s for %s' % (value, key))
                    try:
                        price = Price.objects.get(product=prod, quantity='1')
                    except Price.DoesNotExist:
                        price = Price(product=prod, quantity='1')
                    price.price = value
                    price.save()

            elif opt=="active":
                if value != prod.active:
                    if value:
                        note = "Activated %s"
                    else:
                        note = "Deactivated %s"
                    request.user.message_set.create(message=note % (key))

                    prod.active = value
                    prod.save()

            elif opt=="featured":
                if value != prod.featured:
                    if value:
                        note = "%s is now featured"
                    else:
                        note = "%s is no longer featured"
                    request.user.message_set.create(message=note % (key))

                    prod.featured = value
                    prod.save()

class ProductExportForm(forms.Form):    
    
    def __init__(self, *args, **kwargs):
        products = kwargs.pop('products', None)
        
        super(ProductExportForm, self).__init__(*args, **kwargs)
        
        self.fields['format'] = forms.ChoiceField(label=_('export format'), choices=export_choices(), required=True)
        self.fields['include_images'] = forms.BooleanField(label=_('Include Images'), initial=True, required=False)
        self.fields['include_categories'] = forms.BooleanField(label=_('Include Categories'), initial=True, required=False)
        
        if not products:
            products = Product.objects.by_site().order_by('slug')
            
        for product in products:
            subtypes = product.get_subtypes()
            expclasses = ('export', ) + subtypes
            extclasses = " ".join(expclasses)

            kw = { 
            'label' : product.slug,
            'help_text' : product.name,
            'initial' : False,
            'required' : False,
            'widget' : forms.CheckboxInput(attrs={'class': extclasses}) }
            
            chk = forms.BooleanField(**kw)
            chk.slug = product.slug
            chk.product_id = product.id
            chk.subtypes = " ".join(subtypes)            
            self.fields['export__%s' % product.slug] = chk
            
    def export(self, request):
        self.full_clean()
        format = 'yaml'
        selected = []
        include_images = False
        include_categories = False
        
        for name, value in self.cleaned_data.items():
            if name == 'format':
                format = value
                continue
                
            if name == 'include_images':
                include_images = value
                continue
            
            if name == 'include_categories':
                include_categories = value
                continue
                
            opt, key = name.split('__')
            
            if opt=='export':
                if value:
                    selected.append(key)

        try:
            serializers.get_serializer(format)
        except KeyError:
            raise CommandError("Unknown serialization format: %s" % format)

        objects = []
        images = []
        categories = {}
        for slug in selected:
            product = Product.objects.get(slug=slug)
            objects.append(product)
            for subtype in product.get_subtypes():
                objects.append(getattr(product,subtype.lower()))
            objects.extend(list(product.price_set.all()))
            objects.extend(list(product.productimage_set.all()))
            if include_images:
                for image in product.productimage_set.all():
                    images.append(image.picture)
            if include_categories:
                for category in product.category.all():
                    if category not in categories:
                        categories[category] = 1

        # Export all categories, translations.  Export images,translations if
        # desired.
        if include_categories:
            for category in categories.keys():
                objects.append(category)
                objects.extend(list(category.translations.all()))
                if include_images:
                    for image in category.images.all():
                        objects.append(image)
                        objects.extend(list(image.translations.all()))

        try:
            raw = serializers.serialize(format, objects, indent=False)
        except Exception, e:
            raise CommandError("Unable to serialize database: %s" % e)
            
        if include_images:
            filedir = settings.MEDIA_ROOT
            buf = StringIO()
            zf = zipfile.ZipFile(buf, 'a', zipfile.ZIP_STORED)
            
            export_file = 'products.%s' % format
            zf.writestr(str(export_file), raw)
            
            zinfo = zf.getinfo(str(export_file)) 
            # Caution, highly magic number, chmods the file to 644 
            zinfo.external_attr = 2175008768L 
            
            image_dir = config_value('PRODUCT', 'IMAGE_DIR')
            config = "PRODUCT.IMAGE_DIR=%s\nEXPORT_FILE=%s" % (image_dir, export_file)
            zf.writestr('VARS', config)
            
            zinfo = zf.getinfo('VARS') 
            # Caution, highly magic number, chmods the file to 644 
            zinfo.external_attr = 2175008768L 
            
            for image in images:
                f = os.path.join(filedir, image)
                if os.path.exists(f):
                    zf.write(f, str(image))
            
            zf.close()
            
            raw = buf.getvalue()
            mimetype = "application/zip"
            format = "zip"
        else:
            mimetype = "text/" + format

        response = HttpResponse(mimetype=mimetype, content=raw)
        response['Content-Disposition'] = 'attachment; filename="products-%s.%s"' % (time.strftime('%Y%m%d-%H%M'), format)
            
        return response


class ProductImportForm(forms.Form):  
    
    def __init__(self, *args, **kwargs):        
        super(ProductImportForm, self).__init__(*args, **kwargs)
    
        self.fields['upload'] = forms.Field(label=_("File to import"), widget=forms.FileInput, required=False)

    def import_from(self, infile, maxsize=10000000):
        errors = []
        results = []

        filetype = infile.content_type
        filename = infile.name 
        raw = infile.read()
        
        # filelen = len(raw)
        # if filelen > maxsize:
        #     errors.append(_('Import too large, must be smaller than %i bytes.' % maxsize ))
        
        format = os.path.splitext(filename)[1]
        if format and format.startswith('.'):
            format = format[1:]
        if not format:
            errors.append(_('Could not parse format from filename: %s') % filename)
        
        if format == 'zip':
            zf = zipfile.ZipFile(StringIO(raw), 'r')
            files = zf.namelist()
            image_dir = config_value('PRODUCT', 'IMAGE_DIR')
            other_image_dir = None
            export_file = None
            if 'VARS' in files:
                config = zf.read('VARS')
                lines = [line.split('=') for line in config.split('\n')]
                for key, val in lines:
                    if key == 'PRODUCT.IMAGE_DIR':
                        other_image_dir = val
                    elif key == 'EXPORT_FILE':
                        export_file = val
                
                if other_image_dir is None or export_file is None:
                    errors.append(_('Bad VARS file in import zipfile.'))
                    
                else:
                    # save out all the files which start with other_image_dr
                    rename = image_dir == other_image_dir
                    for f in files:
                        if f.startswith(other_image_dir):
                            buf = zf.read(f)
                            if rename:
                                f = f[len(other_image_dir):]
                                if f[0] in ('/', '\\'):
                                    f = f[1:]
                                f = os.path.join(settings.MEDIA_ROOT, image_dir, f)
                            outf = open(f, 'w')
                            outf.write(buf)
                            outf.close()
                            results.append('Imported image: %s' % f)
                            
                    infile = zf.read(export_file)
                    zf.close()
                    
                    format = os.path.splitext(export_file)[1]
                    if format and format.startswith('.'):
                        format = format[1:]
                    if not format:
                        errors.append(_('Could not parse format from filename: %s') % filename)
                    else:
                        raw = infile
            
            else:
                errors.append(_('Missing VARS in import zipfile.'))
        
        else:
            raw = StringIO(str(raw))

        if not format in serializers.get_serializer_formats():
            errors.append(_('Unknown file format: %s') % format)
            
        if not errors:
            
            from django.db import connection, transaction
            
            transaction.commit_unless_managed()
            transaction.enter_transaction_management()
            transaction.managed(True)
        
            try:

                ct = 0
                models = set()

                for obj in serializers.deserialize(format, raw):
                    obj.save()
                    models.add(obj.object.__class__)
                    ct += 1
                if ct>0:
                    style=no_style()
                    sequence_sql = connection.ops.sequence_reset_sql(style, models)
                    if sequence_sql:
                        cursor = connection.cursor()
                        for line in sequence_sql:
                            cursor.execute(line)
                    
                results.append(_('Added %(count)i objects from %(filename)s') % {'count': ct, 'filename': filename})
                transaction.commit()
                #label_found = True
            except Exception, e:
                #fixture.close()
                errors.append(_("Problem installing fixture '%(filename)s': %(error_msg)s\n") % {'filename': filename, 'error_msg': str(e)})
                errors.append("Raw: %s" % raw)
                transaction.rollback()
                transaction.leave_transaction_management()
            
        return results, errors


class VariationManagerForm(forms.Form):
    dirty = forms.CharField(widget=forms.HiddenInput(), required=False)
    
    def __init__(self, *args, **kwargs):
        self.optionkeys = []
        self.variationkeys = []
        self.existing = {}
        self.optiondict = {}
        self.edit_urls = {}
        self.namedict = {}
        self.skudict = {}
        self.slugdict = {}
        
        self.product = kwargs.pop('product', None)
        
        super(VariationManagerForm, self).__init__(*args, **kwargs)
        
        if self.product:
            configurableproduct = self.product.configurableproduct;
            
            for grp in configurableproduct.option_group.all():
                optchoices = [("%i_%i" % (opt.option_group.id, opt.id), opt.name) for opt in grp.option_set.all()]
                kw = { 
                    'label' : grp.name,
                    'widget' : forms.CheckboxSelectMultiple(),
                    'required' : False,
                    'choices' : optchoices,
                }
                fld = forms.MultipleChoiceField(**kw)
                key = 'optiongroup__%s' % grp.id
                self.fields[key] = fld
                self.optionkeys.append(key)
                self.optiondict[grp.id] = []
                
            configurableproduct.setup_variation_cache()
            for opts in configurableproduct.get_all_options():
                variation = configurableproduct.get_product_from_options(opts)
                optnames = [opt.value for opt in opts]
                kw = { 
                    'initial' : None,
                    'label' : " ".join(optnames),
                    'required' : False
                }
                
                opt_str = '__'.join(["%i_%i" % (opt.option_group.id, opt.id) for opt in opts])    
                
                key = "pv__%s" % opt_str

                if variation:
                    basename = variation.name
                    sku = variation.sku
                    slug = variation.slug
                    kw['initial'] = 'add'
                    self.existing[key] = True
                    self.edit_urls[key] = "/admin/product/product/%i/" % variation.id
                else:
                    basename = u'%s (%s)' % (self.product.name, u'/'.join(optnames))
                    slug = slugify(u'%s_%s' % (self.product.slug, u'_'.join(optnames)))
                    sku = ""

                pv = forms.BooleanField(**kw)

                self.fields[key] = pv
                
                for opt in opts:
                    self.optiondict[opt.option_group.id].append(key)
                
                self.variationkeys.append(key)
                
                # Name Field
                
                nv = forms.CharField(initial=basename)
                namekey = "name__%s" % opt_str
                self.fields[namekey] = nv
                self.namedict[key] = namekey
                
                # SKU Field                

                sv = forms.CharField(initial=sku, required=False)
                skukey = "sku__%s" % opt_str
                self.fields[skukey] = sv
                self.skudict[key] = skukey
                
                # Slug Field
                sf = forms.CharField(initial=slug)
                slugkey = "slug__%s" % opt_str
                self.fields[slugkey] = sf
                self.slugdict[key] = slugkey

    def _save(self, request):
        self.full_clean()
        data = self.cleaned_data
        optiondict = _get_optiondict()
        
        #dirty is a comma delimited list of groupid__optionid strings
        dirty = self.cleaned_data['dirty'].split(',')
        if dirty:
            for key in dirty:
                # this is the keep/create checkbox field
                try:
                    keep = data["pv__" + key]
                    opts = _get_options_for_key(key, optiondict)
                    if opts:
                        if keep:
                            self._create_variation(opts, key, data, request)
                        else:
                            self._delete_variation(opts, request)
                except KeyError:
                    pass
                    
    save = transaction.commit_on_success(_save)
                           
    def _create_variation(self, opts, key, data, request):
        namekey = "name__" + key
        nameval = data[namekey]
        skukey = "sku__" + key
        skuval = data[skukey]
        slugkey = "slug__" + key
        slugval = data[slugkey]
        log.debug("Got name=%s, sku=%s, slug=%s", nameval, skuval, slugval)
        v = self.product.configurableproduct.create_variation(
            opts, 
            name=nameval, 
            sku=skuval, 
            slug=slugval)
        log.info('Updated variation %s', v)
        request.user.message_set.create(message='Created %s' % v)
        return v
        
    def _delete_variation(self, opts, request):
        variation = self.product.configurableproduct.get_product_from_options(opts)
        if variation:
            log.info("Deleting variation for [%s] %s", self.product.slug, opts)
            request.user.message_set.create(message='Deleted %s' % variation)
            variation.delete()
 
def _get_optiondict():
    site = Site.objects.get_current()       
    opts = Option.objects.filter(option_group__site__id = site.id)
    d = {}
    for opt in opts:
        d.setdefault(opt.option_group.id, {})[opt.id] = opt
    return d
    
def _get_options_for_key(key, optiondict):    
    ids = key.split('__')
    opts = []
    for work in ids:
        grpid, optid = work.split('_')
        try:
            opts.append(optiondict[int(grpid)][int(optid)])
        except KeyError:
            log.warn('Could not find option for group id=%s, option id=%s', grpid, optid)
    return opts



from django import template

register = template.Library()

def show_messages(context):
    messages = context['request'].user.get_and_delete_messages()
    return {'visitor_messages' : messages }

register.inclusion_tag('shop/_messages.html', takes_context=True)(show_messages)
